# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

import torch
import ptychi.image_proc as ip
import ptychi.api as api
from ptychi.timing.timer_utils import timer


class PositionCorrection:
    """
    Class containing the various position correction functions used to
    calculate updates to the probe positions.
    """

    def __init__(
        self,
        options: "api.options.base.PositionCorrectionOptions" = None,
    ):
        self.options = options

    @timer()
    def get_update(
        self,
        chi: torch.Tensor,
        obj_patches: torch.Tensor,
        delta_o_patches: torch.Tensor,
        unique_probes: torch.Tensor,
        object_step_size: float,
    ):
        """
        Calculate the position update step direction using the selected position correction function.

        Parameters
        ----------
        chi : torch.Tensor
            A (batch_size, n_modes, h, w) tensor of the exit wave update.
        obj_patches : torch.Tensor
            A (batch_size, n_slices, h, w) tensor of patches of the object. The slice dimension
            is only there to maintain the consistency to the general shape of object patches. 
            Correction algorithms only use the first slice. If position correction should be done
            using other slices, pass the correct slice of the object patches to this function as
            `obj_patches[:, i_slice:i_slice + 1]`.
        delta_o_patches : torch.Tensor
            A (batch_size, n_slices or 1, h, w) tensor of patches of the update to be applied to the object.
        unique_probes : torch.Tensor
            A (batch_size, n_modes, h, w) tensor of unique probes for all positions in the batch.
            The mode dimension is only there to maintain the consistency to the general shape of probe
            patches. Correction algorithms only use the first mode.
        object_step_size : float
            The step size/learning rate of the object optimizer.

        Returns
        -------
        Tensor
            A (n_positions, 2) tensor of updates to the probe positions.
        """
        if self.options.correction_type is api.PositionCorrectionTypes.GRADIENT:
            return self.get_gradient_update(chi, obj_patches, unique_probes)
        elif self.options.correction_type is api.PositionCorrectionTypes.CROSS_CORRELATION:
            return self.get_cross_correlation_update(
                obj_patches, delta_o_patches, unique_probes, object_step_size
            )

    @timer()
    def get_cross_correlation_update(
        self,
        obj_patches: torch.Tensor,
        delta_o_patches: torch.Tensor,
        probe: torch.Tensor,
        object_step_size: float,
    ):
        """
        Use cross-correlation position correction to compute an update to the probe positions.

        Based on the paper:
        - Translation position determination in ptychographic coherent diffraction imaging (2013) - Fucai Zhang
        """
        # Just take the first slice.
        obj_patches = obj_patches[:, 0]
        probe = probe[:, 0]
        delta_o_patches = delta_o_patches[:, 0]
        
        updated_obj_patches = obj_patches + delta_o_patches * object_step_size

        n_positions = len(obj_patches)
        delta_pos = torch.zeros((n_positions, 2))

        probe_thresh = probe.abs().max() * self.options.cross_correlation_probe_threshold
        probe_mask = probe.abs() > probe_thresh

        for i in range(n_positions):
            delta_pos[i] = -ip.find_cross_corr_peak(
                updated_obj_patches[i] * probe_mask[i],
                obj_patches[i] * probe_mask[i],
                scale=self.options.cross_correlation_scale,
                real_space_width=self.options.cross_correlation_real_space_width,
            )

        return delta_pos

    @timer()
    def get_gradient_update(
        self, chi: torch.Tensor, obj_patches: torch.Tensor, probe: torch.Tensor, eps=1e-6
    ):
        """
        Calculate the update direction for probe positions. This routine calculates the gradient with regards
        to probe positions themselves, in contrast to the delta of probe caused by a 1-pixel shift as in
        Odstrcil (2018). However, this is the method implemented in both PtychoShelves and Tike.

        Denote probe positions as s. Given dL/dP = -chi * O.conj() (Eq. 24a), dL/ds = dL/dO * dO/ds =
        real(-chi * P.conj() * grad_O.conj()), where grad_O is the spatial gradient of the probe in x or y.
        """
        # Just take the first slice.
        obj_patches = obj_patches[:, 0]
        
        # Take the first mode of probe and chi.
        probe = probe[:, 0, :, :]
        chi_m0 = chi[:, 0, :, :]
        
        if self.options.differentiation_method == api.ImageGradientMethods.GAUSSIAN:
            dody, dodx = ip.gaussian_gradient(obj_patches, sigma=0.33)
        elif self.options.differentiation_method == api.ImageGradientMethods.FOURIER_DIFFERENTIATION:
            dody, dodx = ip.fourier_gradient(obj_patches)
        elif self.options.differentiation_method == api.ImageGradientMethods.FOURIER_SHIFT:
            dody, dodx = ip.fourier_shift_gradient(obj_patches)
        elif self.options.differentiation_method == api.ImageGradientMethods.NEAREST:
            dody, dodx = ip.nearest_neighbor_gradient(obj_patches, "backward")
        pdodx = dodx * probe
        dldx = (torch.real(pdodx.conj() * chi_m0)).sum(-1).sum(-1)
        denom_x = (pdodx.abs() ** 2).sum(-1).sum(-1)
        dldx = dldx / (denom_x + max(denom_x.max(), eps))

        pdody = dody * probe
        dldy = (torch.real(pdody.conj() * chi_m0)).sum(-1).sum(-1)
        denom_y = (pdody.abs() ** 2).sum(-1).sum(-1)
        dldy = dldy / (denom_y + max(denom_y.max(), eps))

        delta_pos = torch.stack([dldy, dldx], dim=1)

        return delta_pos



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional, Type, Union
import dataclasses
from dataclasses import field


import ptychi.api.options.base as base
import ptychi.api.options.task as task_options
import ptychi.api.enums as enums
import ptychi.api.options.ad_general as ad_general
import ptychi.forward_models as fm


@dataclasses.dataclass
class AutodiffPtychographyReconstructorOptions(ad_general.AutodiffReconstructorOptions):
    forward_model_class: Union["enums.ForwardModels", Type["fm.ForwardModel"]] = enums.ForwardModels.PLANAR_PTYCHOGRAPHY
    
    def get_reconstructor_type(self) -> enums.Reconstructors:
        return enums.Reconstructors.AD_PTYCHO


@dataclasses.dataclass
class DeepImagePriorOptions(base.Options):
    enabled: bool = False
    """Whether to use deep image prior."""
    
    model: enums.DIPModels = enums.DIPModels.UNET
    """The model used to represent the reconstruction parameter."""
    
    model_params: dict = dataclasses.field(default_factory=dict)
    """The parameters for the model's constructor."""
    
    net_input_channels: int = 32
    """The number of channels in the input to the network."""
    
    constrain_object_outside_network: bool = False
    """If True, magnitude and phase of the object generated by the model is
    further constrained so that the magnitude is between 0 and 1 outside the
    network model. Do not enable this option if your model already includes
    such constraints.
    """
    
    residual_generation: bool = False
    """If True, data generated by the DIP network is added to the initial guess
    of the reconstruction parameter, i.e., the network learns a residue. When
    using this feature, the network is usually supposed to start from a state
    that maps everything to zeros, so that the overall DIP generation is an
    identity function.
    """
    
    
@dataclasses.dataclass
class AutodiffPtychographyObjectExperimentalOptions(base.Options):
    deep_image_prior_options: Optional[DeepImagePriorOptions] = dataclasses.field(default_factory=DeepImagePriorOptions)


@dataclasses.dataclass
class AutodiffPtychographyObjectOptions(base.ObjectOptions):
    experimental: AutodiffPtychographyObjectExperimentalOptions = dataclasses.field(default_factory=AutodiffPtychographyObjectExperimentalOptions)


@dataclasses.dataclass
class AutodiffPtychographyProbeExperimentalOptions(base.Options):
    deep_image_prior_options: Optional[DeepImagePriorOptions] = dataclasses.field(default_factory=DeepImagePriorOptions)


@dataclasses.dataclass
class AutodiffPtychographyProbeOptions(base.ProbeOptions):
    experimental: AutodiffPtychographyProbeExperimentalOptions = dataclasses.field(default_factory=AutodiffPtychographyProbeExperimentalOptions)


@dataclasses.dataclass
class AutodiffPtychographyProbePositionOptions(base.ProbePositionOptions):
    pass


@dataclasses.dataclass
class AutodiffPtychographyOPRModeWeightsOptions(base.OPRModeWeightsOptions):
    pass


@dataclasses.dataclass
class AutodiffPtychographyOptions(task_options.PtychographyTaskOptions):
    reconstructor_options: AutodiffPtychographyReconstructorOptions = field(
        default_factory=AutodiffPtychographyReconstructorOptions
    )

    object_options: AutodiffPtychographyObjectOptions = field(
        default_factory=AutodiffPtychographyObjectOptions
    )

    probe_options: AutodiffPtychographyProbeOptions = field(
        default_factory=AutodiffPtychographyProbeOptions
    )

    probe_position_options: AutodiffPtychographyProbePositionOptions = field(
        default_factory=AutodiffPtychographyProbePositionOptions
    )

    opr_mode_weight_options: AutodiffPtychographyOPRModeWeightsOptions = field(
        default_factory=AutodiffPtychographyOPRModeWeightsOptions
    )



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

import torch
import torch.nn as nn
import torch.nn.functional as F

from ptychi.reconstructors.nn.components import DoubleConv


class Down(nn.Module):
    """Downscaling with maxpool then double conv.
    
    This implementation is adapted from https://github.com/milesial/Pytorch-UNet.
    """

    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_channels, out_channels)
        )

    def forward(self, x):
        return self.maxpool_conv(x)


class Up(nn.Module):
    """Upscaling then double conv.
    
    This implementation is adapted from https://github.com/milesial/Pytorch-UNet.
    """

    def __init__(
        self, 
        in_channels: int, 
        out_channels: int, 
        bilinear: bool = True, 
        skip_connection: bool = True,
    ):
        super().__init__()
        self.skip_connection = skip_connection

        # if bilinear, use the normal convolutions to reduce the number of channels
        up_out_channels = in_channels // 2 if skip_connection else in_channels
        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
            self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)
        else:
            self.up = nn.ConvTranspose2d(in_channels, up_out_channels, kernel_size=2, stride=2)
            self.conv = DoubleConv(in_channels, out_channels)

    def forward(self, x1, x2):
        """The forward pass.

        Parameters
        ----------
        x1 : Tensor
            The output of the previous layer.
        x2 : Tensor
            The output of the layer in the downsampling part that is on the same scale.
            When skip connection is enabled, this tensor is concatenated with the output.
            Even when skip connection is disabled, this argument is still required because
            it is used to determine the padding of the input tensor.

        Returns
        -------
        Tensor
            The output of the layer.
        """
        x1 = self.up(x1)
        
        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2,
                        diffY // 2, diffY - diffY // 2])
        # if you have padding issues, see
        # https://github.com/HaiyongJiang/U-Net-Pytorch-Unstructured-Buggy/commit/0e854509c2cea854e247a9c615f175f76fbb2e3a
        # https://github.com/xiaopeng-liao/Pytorch-UNet/commit/8ebac70e633bac59fc22bb5195e513d5832fb3bd
        if self.skip_connection:
            x = torch.cat([x2, x1], dim=1)
        else:
            x = x1
        return self.conv(x)


class UNet(nn.Module):
    def __init__(
        self, 
        num_in_channels: int, 
        num_out_channels: int, 
        bilinear: bool = True, 
        initialize: bool = False,
        skip_connections: bool = True,
    ):
        """U-net model based on 
        Ronneberger, O., Fischer, P., & Brox, T. (2015). U-Net: Convolutional Networks for 
        Biomedical Image Segmentation. arXiv:1505.04597.
        
        This implementation is adapted from https://github.com/milesial/Pytorch-UNet.

        Parameters
        ----------
        n_channels : int
            Number of channels in the input image.
        bilinear : bool, optional
            If True, use bilinear upsampling. Otherwise, use transposed convolutions.
        initialize : bool, optional
            If True, initialize the model with normal weights.
        skip_connections : bool, optional
            If True, use skip connections.
        """
        super(UNet, self).__init__()
        self.n_in_channels = num_in_channels
        self.bilinear = bilinear

        self.inc = DoubleConv(num_in_channels, 64)
        self.down1 = Down(64, 128)
        self.down2 = Down(128, 256)
        self.down3 = Down(256, 512)
        factor = 2 if bilinear else 1
        self.down4 = Down(512, 1024 // factor)
        if skip_connections:
            self.up1 = Up(1024, 512 // factor, bilinear, skip_connection=skip_connections)
            self.up2 = Up(512, 256 // factor, bilinear, skip_connection=skip_connections)
            self.up3 = Up(256, 128 // factor, bilinear, skip_connection=skip_connections)
            self.up4 = Up(128, 64, bilinear, skip_connection=skip_connections)
            self.outc = nn.Conv2d(64, num_out_channels, kernel_size=1)
        else:
            self.up1 = Up(1024 // factor, 512 // factor, bilinear, skip_connection=skip_connections)
            self.up2 = Up(512 // factor, 256 // factor, bilinear, skip_connection=skip_connections)
            self.up3 = Up(256 // factor, 128 // factor, bilinear, skip_connection=skip_connections)
            self.up4 = Up(128 // factor, 64, bilinear, skip_connection=skip_connections)
            self.outc = nn.Conv2d(64, num_out_channels, kernel_size=1)
        
        if initialize:
            self.apply(self._initialize_weights)
            
    def _initialize_weights(self, m):
        if isinstance(m, nn.Conv2d):
            nn.init.normal_(m.weight, 0, 1e-2)
            if m.bias is not None:
                nn.init.normal_(m.bias, 0, 1e-2)

    def forward(self, x):
        """Forward pass of the U-net model.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape (batch_size, n_in_channels, height, width).
        
        Returns
        -------
        torch.Tensor
            Output tensor of shape (batch_size, n_out_channels, height, width).
        """
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        logits = self.outc(x)
        return logits

    def use_checkpointing(self):
        self.inc = torch.utils.checkpoint(self.inc)
        self.down1 = torch.utils.checkpoint(self.down1)
        self.down2 = torch.utils.checkpoint(self.down2)
        self.down3 = torch.utils.checkpoint(self.down3)
        self.down4 = torch.utils.checkpoint(self.down4)
        self.up1 = torch.utils.checkpoint(self.up1)
        self.up2 = torch.utils.checkpoint(self.up2)
        self.up3 = torch.utils.checkpoint(self.up3)
        self.up4 = torch.utils.checkpoint(self.up4)
        self.outc = torch.utils.checkpoint(self.outc)



import random
import os
import datetime
import logging
import configparser
from typing import Union

import torch
import h5py
import numpy as np
import pytest
import subprocess
import socket

from ptychi.utils import rescale_probe, add_additional_opr_probe_modes_to_probe, set_default_complex_dtype, to_tensor
import ptychi.timing.timer_utils as timer_utils
from ptychi.timing.io import save_timing_data_to_unique_file_path


class BaseTester:
    def setup_method(
        self,
        name="",
        generate_data=False,
        generate_gold=False,
        debug=False,
        action=None,
        pytestconfig=None,
    ):
        """
        A Pytest hook that sets instance attributes before running each test method. 
        If the script is executed with `python`, this method will not run automatically
        before calling a test method. Therefore, it can be used to set instance attributes
        for all methods in a code snippet if that snippet is intended to be executed
        with `python`.

        Parameters
        ----------
        name : str, optional
            The name of the tester.
        generate_data : bool
            Whether to generate test data. 
        generate_gold : bool
            Whether to generate gold data. 
        save_timing : bool
            Whether to save timing results.
        debug : bool, optional
            Switches debug mode.
        """
        logging.basicConfig(level=logging.INFO)
        
        self.name = name
        
        self.generate_data = generate_data
        self.generate_gold = generate_gold
        # self.save_timing = save_timing
        self.debug = debug
        
        if pytestconfig is not None:
            self.high_tol = pytestconfig.getoption("high_tol")
            self.action = pytestconfig.getoption("action")
            self.save_timing = pytestconfig.getoption("save_timing")
        else:
            self.high_tol = False
            self.action = action
            self.save_timing = False
            
        self.atol = 1e-3
        self.rtol = 1e-2 if self.high_tol else 1e-4
    
    @pytest.fixture(autouse=True)
    def inject_config(self, pytestconfig):
        self.pytestconfig = pytestconfig
        self.setup_method(
            name="",
            generate_data=False,
            generate_gold=False,
            debug=False,
            action=None,
            pytestconfig=pytestconfig,
        )
    
    @staticmethod
    def get_ci_data_dir():
        try:
            dir = os.environ['PTYCHO_CI_DATA_DIR']
        except KeyError:
            raise KeyError('PTYCHO_CI_DATA_DIR not set. Please set it to the path to the data folder.')
        return dir

    def get_ci_input_data_dir(self):
        return os.path.join(self.get_ci_data_dir(), 'data')

    def get_ci_gold_data_dir(self):
        return os.path.join(self.get_ci_data_dir(), 'gold_data')
    
    """
    @property
    def gold_run_id(self) -> int:
        return int(self.ci_config['TESTING']['GoldRunID'])
        
    def get_run_id(self) -> int:
        \"""
        Get the current run ID as the maximum of the existing run IDs in the output directory plus 1.
        If the output directory does not contain any run IDs, return 1.

        Returns
        -------
        int
            The current run ID.
        \"""
        output_dir = self.get_output_dir()
        run_ids = [int(f.name) for f in os.scandir(output_dir) if f.is_dir()]
        if len(run_ids) == 0:
            return 1
        return max(run_ids) + 1
        
    @staticmethod
    def get_input_dir():
        try:
            dir = os.environ['PTYCHI_CI_INPUT_DIR']
        except KeyError:
            raise KeyError('PTYCHI_CI_INPUT_DIR not set. Please set it to the path to the data folder.')
        return dir
    
    @staticmethod
    def get_output_dir():
        try:
            dir = os.environ['PTYCHI_CI_OUTPUT_DIR']
        except KeyError:
            raise KeyError('PTYCHI_CI_OUTPUT_DIR not set. Please set it to the path to the data folder.')
        return dir
    
    def get_ci_config(self):
        try:
            config_ini = os.environ['PTYCHI_CI_CONFIG_PATH']
        except KeyError:
            raise KeyError('PTYCHI_CI_CONFIG_PATH not set. Please set it to the path to the config.ini file.')
        config = configparser.ConfigParser()
        config.read(config_ini)
        return config
    """
    
    def save_gold_data(self, name, data):
        fname = os.path.join(self.get_ci_gold_data_dir(), name, 'recon.npy')
        if not os.path.exists(os.path.dirname(fname)):
            os.makedirs(os.path.dirname(fname))
        np.save(fname, data)
    
    def load_gold_data(self, name):
        fname = os.path.join(self.get_ci_gold_data_dir(), name, 'recon.npy')
        return np.load(fname)
    
    def run_comparison(self, name, test_data, atol=None, rtol=None):
        if atol is None:
            atol = self.atol
        if rtol is None:
            rtol = self.rtol
        gold_data = self.load_gold_data(name)
        compare_data(test_data, gold_data, atol=atol, rtol=rtol, name=name)
        return
    
    def get_default_input_data_file_paths(self, name):
        dp = os.path.join(self.get_ci_input_data_dir(), name, 'ptychodus_dp.hdf5')
        para = os.path.join(self.get_ci_input_data_dir(), name, 'ptychodus_para.hdf5')
        return dp, para

    def setup_ptychi(self, cpu_only=False, gpu_indices=(0,)):
        torch.manual_seed(123)
        random.seed(123)
        np.random.seed(123)
        
        torch.set_default_device('cpu' if cpu_only else 'cuda')
        torch.set_default_dtype(torch.float32)
        set_default_complex_dtype(torch.complex64)
        
        if not cpu_only:
            os.environ['CUDA_VISIBLE_DEVICES'] = ','.join(map(str, gpu_indices))

    def load_data_ptychodus(self, diffraction_pattern_file, parameter_file, subtract_position_mean=False, additional_opr_modes=0, scale_probe_magnitude=True):
        patterns = h5py.File(diffraction_pattern_file, 'r')['dp'][...]

        f_meta = h5py.File(parameter_file, 'r')
        probe = f_meta['probe'][...]
        if scale_probe_magnitude:
            probe = rescale_probe(probe, patterns)
        if probe.ndim == 3:
            probe = probe[None, :, :, :]
        probe = to_tensor(probe)
        if additional_opr_modes > 0:
            probe = add_additional_opr_probe_modes_to_probe(probe, n_opr_modes_to_add=additional_opr_modes)
        
        positions = np.stack([f_meta['probe_position_y_m'][...], f_meta['probe_position_x_m'][...]], axis=1)
        pixel_size_m = f_meta['object'].attrs['pixel_height_m']
        positions_px = positions / pixel_size_m
        if subtract_position_mean:
            positions_px -= positions_px.mean(axis=0)
        
        return patterns, probe, pixel_size_m, positions_px
    
    def plot_object(self, obj):
        if obj.ndim == 2:
            obj = obj[None, ...]
        if obj.shape[0] > 1:
            plot_multislice_phase(obj)
        else:
            plot_complex_image(obj)
    
    @staticmethod
    def wrap_recon_tester(name="", run_comparison=True):
        """
        A decorator factory that wraps a test method to generate or compare data.

        Parameters
        ----------
        name : str, optional
            The name of the test.
        """
        def decorator(test_method):
            def wrapper(self: BaseTester):
                if self.save_timing:
                    timer_utils.toggle_timer(enable=True)
                recon = test_method(self)
                if self.debug and not self.generate_gold:
                    self.plot_object(recon)
                if self.generate_gold:
                    self.save_gold_data(name, recon)
                if not self.generate_gold and run_comparison:
                    self.run_comparison(name, recon, atol=self.atol, rtol=self.rtol)
                if self.save_timing:
                    save_timing_data_to_unique_file_path(name, get_timing_data_dir())
            return wrapper
        return decorator
    

class TungstenDataTester(BaseTester):
    
    def load_tungsten_data(self, additional_opr_modes=0, pos_type='true'):
        return self.load_data_ptychodus(
            os.path.join(self.get_ci_input_data_dir(), '2d_ptycho', 'dp_250.hdf5'), 
            os.path.join(self.get_ci_input_data_dir(), '2d_ptycho', 'metadata_250_{}Pos.hdf5'.format(pos_type)), 
            additional_opr_modes=additional_opr_modes
        )


def get_timestamp():
    return datetime.datetime.now().strftime('%Y%m%d%H%M')


def compare_data(test_data, gold_data, atol=1e-7, rtol=1e-7, name=""):
    if not np.allclose(gold_data.shape, test_data.shape):
        print('{} FAILED [SHAPE MISMATCH]'.format(name))
        print('  Gold shape: {}'.format(gold_data.shape))
        print('  Test shape: {}'.format(test_data.shape))
        raise AssertionError
    if not np.allclose(gold_data, test_data, atol=atol, rtol=rtol):
        print('{} FAILED [MISMATCH]'.format(name))
        abs_diff = np.abs(gold_data - test_data)
        loc_max_diff = np.unravel_index(np.argmax(abs_diff), abs_diff.shape)
        loc_max_diff = [a.item() for a in loc_max_diff]
        print('  Mean abs diff: {}'.format(abs_diff.mean()))
        print('  Location of max diff: {}'.format(loc_max_diff))
        print('  Max abs diff: {}'.format(abs_diff[*loc_max_diff].item()))
        print('  Value at max abs diff (test): {}'.format(test_data[*loc_max_diff].item()))
        print('  Value at max abs diff (gold): {}'.format(gold_data[*loc_max_diff].item()))
        raise AssertionError
    print('{} PASSED'.format(name))


def plot_complex_image(img):
    import matplotlib.pyplot as plt
    if img.ndim == 3:
        img = img[0]
    fig, ax = plt.subplots(1, 2)
    ax[0].imshow(np.abs(img))
    ax[0].set_title('magnitude')
    ax[1].imshow(np.angle(img))
    ax[1].set_title('phase')
    plt.show()


def plot_multislice_phase(img):
    import matplotlib.pyplot as plt
    n_slices = img.shape[0]
    fig, ax = plt.subplots(n_slices, 1)
    for i in range(n_slices):
        ax[i].imshow(np.angle(img[i, ...]))
        ax[i].set_title('slice {}'.format(i))
    plt.show()
    
    
def plot_probe_positions(pos, pos_true=None):
    import matplotlib.pyplot as plt
    plt.scatter(pos[:, 1], pos[:, 0], label="reconstructed")
    if pos_true is not None:
        plt.scatter(pos_true[:, 1], pos_true[:, 0], label="true")
    plt.legend()
    plt.show()


def get_timing_data_dir():
    return os.path.join(BaseTester.get_ci_data_dir(), "timing")


# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Literal, Optional, Tuple, Union, Callable

import torch
import numpy as np

from ptychi.timing.timer_utils import timer

_use_double_precision_for_fft = True
_allow_nondeterministic_algorithms = False


@timer()
def trim_mean(
    x: torch.Tensor, 
    fraction: float = 0.1, 
    dim: Optional[Union[int, Tuple[int, ...]]] = None,
    keepdim: bool = False
) -> torch.Tensor:
    """
    Calculate the mean of a tensor after removing a certain percentage of the
    lowest and highest values.

    Parameters
    ----------
    x : Tensor
        Input tensor.
    fraction : float, optional
        The fraction of trim, between 0 and 1. Default is 0.1.
    dim : int or tuple of int, optional
        The axis/axes along which to calculate the mean.

    Returns
    -------
    trimmed_mean : tensor
        The trimmed mean.
    """
    lb = torch.quantile(x, fraction, dim=dim, keepdim=True)
    ub = torch.quantile(x, 1 - fraction, dim=dim, keepdim=True)
    mask = (x >= lb) & (x <= ub)

    mask_check = torch.sum(mask, dim=dim)
    mask_check = torch.all(mask_check > 0)
    if mask_check:
        x = x.clone()
        x[~mask] = torch.nan
        return torch.nanmean(x, dim=dim, keepdim=keepdim)
    else:
        return torch.mean(x, dim=dim, keepdim=keepdim)
    
    
def get_use_double_precision_for_fft():
    return _use_double_precision_for_fft


def set_use_double_precision_for_fft(use_double_precision_for_fft: bool):
    global _use_double_precision_for_fft
    _use_double_precision_for_fft = use_double_precision_for_fft
    
    
def get_allow_nondeterministic_algorithms():
    return _allow_nondeterministic_algorithms


def set_allow_nondeterministic_algorithms(allow: bool):
    global _allow_nondeterministic_algorithms
    _allow_nondeterministic_algorithms = allow


def angle(x: torch.Tensor, eps=1e-5) -> torch.Tensor:
    """
    A stable version of `torch.angle()`, which calculates
    arctan(imag / (real + eps)) to prevent unstable results when
    the real part is small (so that the ratio dangles between -inf
    and inf).

    Parameters
    ----------
    x : torch.Tensor
        The complex tensor as input.
    eps : float
        The stablization constant.

    Returns
    -------
    torch.Tensor
        The phase of the complex tensor.
    """
    return torch.atan2(x.imag, x.real + eps)


@timer()
def orthogonalize_gs(
    x: torch.Tensor,
    dim: Union[int, Tuple[int, ...]] = -1,
    group_dim: Union[int, None] = None,
) -> torch.Tensor:
    """
    Gram-schmidt orthogonalization for complex arrays. Adapted from
    Tike (https://github.com/AdvancedPhotonSource/tike).

    Parameters
    ----------
    x : Tensor
        Data to be orthogonalized.
    dim : int or tuple of int, optional
        The axis/axes to be orthogonalized. By default only the last axis is
        orthogonalized. If axis is a tuple, then the number of orthogonal
        vectors is the length of the last dimension not included in axis. The
        other dimensions are broadcast.
    group_dim : int, optional
        The axis along which to orthogonalize. Other dimensions are broadcast.

    Returns
    -------
    Tensor
        Orthogonalized data.
    """
    x, dim, group_dim = _prepare_data_for_orthogonalization(x, dim, group_dim, move_group_dim_to=0)
    u = x.clone()
    for i in range(1, len(x)):
        u[i:] -= project(x[i:], u[i - 1 : i], dim=dim)
    return torch.moveaxis(u, 0, group_dim)


@timer()
def orthogonalize_svd(
    x: torch.Tensor,
    dim: Union[int, Tuple[int, ...]] = -1,
    group_dim: Union[int, None] = None,
    preserve_norm: bool = False,
) -> torch.Tensor:
    """
    SVD orthogonalization for complex arrays. Adapted from PtychoShelves (probe_modes_ortho.m).

    Parameters
    ----------
    x : Tensor
        Data to be orthogonalized.
    dim : int or tuple of int, optional
        The axis/axes to be orthogonalized. By default only the last axis is
        orthogonalized. If axis is a tuple, then the number of orthogonal
        vectors is the length of the last dimension not included in axis. The
        other dimensions are broadcast.
    group_dim : int, optional
        The axis along which to orthogonalize. Other dimensions are broadcast.

    Returns
    -------
    Tensor
        Orthogonalized data.
    """
    x, dim, group_dim = _prepare_data_for_orthogonalization(
        x, dim, group_dim, move_group_dim_to=None
    )
    if preserve_norm:
        orig_norm = norm(x, dim=list(dim) + [group_dim], keepdims=True)

    # Move group_dim to the end, and move dim right before group_dim.
    x = torch.moveaxis(x, list(sorted(dim)) + [group_dim], list(range(-len(dim) - 1, 0)))
    group_dim_shape = [x.shape[-1]]
    dim_shape = list(x.shape[-(len(dim) + 1) : -1])
    batch_dim_shape = list(x.shape[: -(len(dim) + 1)])

    # Now the axes of x is rearranged to (bcast_dims, *dim, group_dim).
    # Straighten dimensions given by `dim`.`
    # Shape of x:        (bcast_dims_shape, prod(dim_shape), group_dim_shape)
    x = x.reshape(batch_dim_shape + [-1] + group_dim_shape)
    
    # Use higher precision.
    orig_dtype = x.dtype
    if orig_dtype.is_complex:
        x = x.type(torch.complex128)
    else:
        x = x.type(torch.float64)

    # Creat an shape(group_dim) x shape(group_dim) covariance matrix and eigendecompose it.
    if x.ndim == 2:
        covmat = x.transpose(-1, -2) @ x.conj()
    else:
        covmat = torch.bmm(x.transpose(-1, -2), x.conj())

    evals, evecs = torch.linalg.eig(covmat)

    sorted_inds = torch.argsort(evals.abs(), dim=-1, descending=True)
    # Shape of evecs:    (bcast_dims_shape, group_dim_shape, group_dim_shape)
    evecs = torch.gather(
        evecs,
        dim=-1,
        index=sorted_inds[..., None, :].repeat(*([1] * (evecs.ndim - 2)), evecs.shape[-2], 1),
    )

    if x.ndim == 2:
        x = x @ evecs.conj()
    else:
        x = torch.bmm(x, evecs.conj())

    # Restore u to the original shape.
    # Unflatten `dim`.
    x = x.reshape(batch_dim_shape + dim_shape + group_dim_shape)
    # Then, Move `group_dim` and `dim` back.
    x = torch.moveaxis(x, list(range(-len(dim) - 1, 0)), list(sorted(dim)) + [group_dim])

    if preserve_norm:
        new_norm = norm(x, dim=list(dim) + [group_dim], keepdims=True)
        x = x * (orig_norm / new_norm)

    return x.type(orig_dtype)


def project(a, b, dim=None):
    """Return complex vector projection of a onto b for along given axis."""
    projected_length = inner(a, b, dim=dim, keepdims=True) / inner(b, b, dim=dim, keepdims=True)
    return projected_length * b

def inner(x, y, dim=None, keepdims=False):
    """Return the complex inner product; the order of the operands matters."""
    return (x * y.conj()).sum(dim, keepdims=keepdims)


def mnorm(x, dim=-1, keepdims=False):
    """Return the vector 2-norm of x but replace sum with mean."""
    return torch.sqrt(torch.mean((x * x.conj()).real, dim=dim, keepdims=keepdims))


def norm(x, dim=-1, keepdims=False):
    """Return the vector 2-norm of x along given axis."""
    return torch.sqrt(torch.sum((x * x.conj()).real, dim=dim, keepdims=keepdims))


def _prepare_data_for_orthogonalization(
    x: torch.Tensor,
    dim: Union[int, Tuple[int, ...]] = -1,
    group_dim: Union[int, None] = None,
    move_group_dim_to: Optional[Union[int, Literal["before_dim"]]] = None,
) -> Tuple[torch.Tensor, Tuple[int, ...], int]:
    """
    Find `dim` and `group_dim`.
    Permute the axes of x, such that group_dim is at the front.
    """
    # If dim contains negative indices, convert them to positive indices.
    try:
        dim = tuple(a % x.ndim for a in dim)
    except TypeError:
        dim = (dim % x.ndim,)
    # Find group_dim, the last dimension not included in axis; we iterate over N
    # vectors in the Gram-schmidt algorithm. Dimensions that are not N or
    # included in axis are leading dimensions for broadcasting.
    if group_dim is None:
        group_dim = x.ndim - 1
        while group_dim in dim:
            group_dim -= 1
    group_dim = group_dim % x.ndim
    if group_dim in dim:
        raise ValueError("Cannot orthogonalize a single vector.")
    if move_group_dim_to is not None:
        if move_group_dim_to == "before_dim":
            x = torch.moveaxis(x, group_dim, min(dim) - 1)
        else:
            x = torch.moveaxis(x, group_dim, move_group_dim_to)
    return x, dim, group_dim


def polyfit(x: torch.Tensor, y: torch.Tensor, deg: int = 1):
    """
    Fit a polynomial to the data and return the coefficients
    from high to low order.
    
    Parameters
    ----------
    x : torch.Tensor
        The independent variable.
    y : torch.Tensor
        The dependent variable.
    deg : int
        The degree of the polynomial to fit.
    
    Returns
    -------
    torch.Tensor
        The coefficients of the polynomial.
    """
    x_powers = x[:, None] ** torch.arange(deg, -1, -1)
    return torch.linalg.lstsq(x_powers, y, rcond=None)[0]


def polyval(x: torch.Tensor, coeffs: torch.Tensor):
    """
    Evaluate a polynomial at the given points.
    
    Parameters
    ----------
    x : torch.Tensor
        The independent variable.
    coeffs : torch.Tensor
        The coefficients of the polynomial from high to low order.
    
    Returns
    -------
    torch.Tensor
        The values of the polynomial at the given points.
    """
    return (coeffs * x.view(-1, 1) ** torch.arange(len(coeffs) - 1, -1, -1)).sum(1)


def real_dtype_to_complex(dtype: torch.dtype) -> torch.dtype:
    """Convert a real dtype to a complex dtype with the same precision.
    If a complex dtype is provided, it is returned unchanged.

    Parameters
    ----------
    dtype : torch.dtype
        The real dtype to convert.

    Returns
    -------
    torch.dtype
        The complex dtype with the same precision.
    """
    if dtype.is_complex:
        return dtype
    elif dtype == torch.float64:
        return torch.complex128
    elif dtype == torch.float32:
        return torch.complex64
    elif dtype == torch.float16:
        return torch.complex32
    else:
        raise ValueError(f"Unsupported dtype: {dtype}")
    

def complex_dtype_to_real(dtype: torch.dtype) -> torch.dtype:
    """Convert a complex dtype to a real dtype with the same precision.
    If a real dtype is provided, it is returned unchanged.

    Parameters
    ----------
    dtype : torch.dtype
        The complex dtype to convert.

    Returns
    -------
    torch.dtype
        The real dtype with the same precision.
    """
    if not dtype.is_complex:
        return dtype
    elif dtype == torch.complex128:
        return torch.float64
    elif dtype == torch.complex64:
        return torch.float32
    elif dtype == torch.complex32:
        return torch.float16
    else:
        raise ValueError(f"Unsupported dtype: {dtype}")


def fft2_precise(x, norm=None):
    """
    2D FFT with double precision.
    """
    if get_use_double_precision_for_fft():
        final_dtype = real_dtype_to_complex(x.dtype)
        return torch.fft.fft2(x.type(torch.complex128), norm=norm).type(final_dtype)
    else:
        return torch.fft.fft2(x, norm=norm)


def ifft2_precise(x, norm=None):
    """
    2D FFT with double precision.
    """
    if get_use_double_precision_for_fft():
        final_dtype = real_dtype_to_complex(x.dtype)
        return torch.fft.ifft2(x.type(torch.complex128), norm=norm).type(final_dtype)
    else:
        return torch.fft.ifft2(x, norm=norm)


def is_all_integer(x: torch.Tensor) -> bool:
    """Check if all elements in a tensor are integers."""
    return torch.allclose(torch.eq(x, torch.round(x)), atol=1e-7)


def masked_argmax(
    x: torch.Tensor | np.ndarray, 
    mask: torch.Tensor | np.ndarray
) -> torch.Tensor | np.ndarray:
    """
    Find the index of the maximum value in a tensor within a mask.
    
    This function works for both torch and numpy arrays.
    
    Parameters
    ----------
    x : torch.Tensor | np.ndarray
        The input tensor.
    mask : torch.Tensor | np.ndarray
        The mask. Search is only performed within the masked region.
    
    Returns
    -------
    torch.Tensor | np.ndarray
        The index of the maximum value in the tensor within the mask.
    """
    if isinstance(x, torch.Tensor):
        x = torch.where(mask, x, -torch.inf)
        return torch.argmax(x)
    else:
        x = np.where(mask, x, -np.inf)
        return np.argmax(x)


def masked_argmin(
    x: torch.Tensor | np.ndarray, 
    mask: torch.Tensor | np.ndarray
) -> torch.Tensor | np.ndarray:
    """
    Find the index of the minimum value in a tensor within a mask.
    
    This function works for both torch and numpy arrays.
    
    Parameters
    ----------
    x : torch.Tensor | np.ndarray
        The input tensor.
    mask : torch.Tensor | np.ndarray
        The mask. Search is only performed within the masked region.
    
    Returns
    -------
    torch.Tensor | np.ndarray
        The index of the minimum value in the tensor within the mask.
    """
    if isinstance(x, torch.Tensor):
        x = torch.where(mask, x, torch.inf)
        return torch.argmin(x)
    else:
        x = np.where(mask, x, np.inf)
        return np.argmin(x)


def decompose_2x2_affine_transform_matrix(
    m: torch.Tensor
) -> torch.Tensor:
    """Decompose a 2x2 affine transformation matrix (without translation) into 
    scale, asymmetry, rotation, and shear.
    
    Let x1, x2, x3, x4 be the scale, asymmetry, rotation, and shear. The affine
    transformation matrix is decomposed into
    x1 * ((1+x2/2, 0     )  * ((cos x3,  sin x3)  * ((1,      0)
          (0,      1-x2/2))    (-sin x3, cos x3))    (tan x4, 1))
    and the 
    """
    def f(x):
        y = torch.stack([
            (1 + x[1] / 2) * (torch.cos(x[2]) + torch.sin(x[2]) * torch.tan(x[3])),
            (1 + x[1] / 2) * torch.sin(x[2]),
            (1 - x[1] / 2) * (-torch.sin(x[2]) + torch.cos(x[2]) * torch.tan(x[3])),
            (1 - x[1] / 2) * torch.cos(x[2])
        ]) * x[0]
        return y
    x = nonlinear_lsq(torch.tensor([1.0, 0.0, 0.0, 0.0]), m.reshape(-1), f)
    return x


def compose_2x2_affine_transform_matrix(
    scale: torch.Tensor,
    asymmetry: torch.Tensor,
    rotation: torch.Tensor,
    shear: torch.Tensor
) -> torch.Tensor:
    """
    Compose a 2x2 affine transformation matrix from scale, asymmetry, rotation, and shear.
    """
    x = [scale, asymmetry, rotation, shear]
    m = torch.stack([
        (1 + x[1] / 2) * (torch.cos(x[2]) + torch.sin(x[2]) * torch.tan(x[3])),
        (1 + x[1] / 2) * torch.sin(x[2]),
        (1 - x[1] / 2) * (-torch.sin(x[2]) + torch.cos(x[2]) * torch.tan(x[3])),
        (1 - x[1] / 2) * torch.cos(x[2])
    ]) * x[0]
    return m.reshape(2, 2)


def nonlinear_lsq(
    x0: torch.Tensor,
    y: torch.Tensor,
    f: Callable[[torch.Tensor], torch.Tensor],
    n_iter: int = 20,
    **kwargs
) -> torch.Tensor:
    """Solve a nonlinear least squares problem.
    
    Parameters
    ----------
    x0 : torch.Tensor
        The initial guess for the solution.
    y : torch.Tensor
        The data.
    f : Callable[torch.Tensor, torch.Tensor]
        The function to minimize.
    """
    x = x0.clone().requires_grad_(True)
    optimizer = torch.optim.LBFGS([x], **kwargs)
    loss_history = []
    
    def compute_loss():
        return ((y - f(x)).abs() ** 2).sum()
    
    def closure():
        optimizer.zero_grad()
        loss = compute_loss()
        loss.backward()
        return loss
    
    for _ in range(n_iter):
        loss_history.append(compute_loss().item())
        optimizer.step(closure)
    
    # Convergence check.
    loss_history = torch.tensor(loss_history)
    window_size = max(len(loss_history) // 10, 1)
    avg_loss_1 = loss_history[-window_size:].mean()
    avg_loss_2 = loss_history[-2 * window_size:-window_size].mean()
    if torch.abs(avg_loss_2) > 0 and avg_loss_1 / avg_loss_2 > 1.1:
        raise ValueError("Non-linear least squares did not converge.")
    return x.detach()


def fit_linear_transform_matrix(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """Fit a linear transformation matrix A, so that y = A @ x.
    
    Parameters
    ----------
    x : torch.Tensor
        A (n_points, n_features_in) tensor of input data.
    y : torch.Tensor
        A (n_points, n_features_out) tensor of output data.
    
    Returns
    -------
    torch.Tensor
        The linear transformation matrix.
    """
    a_mat = torch.linalg.lstsq(x, y, rcond=None)[0]
    return a_mat

  
def reprod(a, b):
    """
    Real part of the product of 2 arrays
    """
    return a.real * b.real + a.imag * b.imag


def redot(a, b, axis=None):
    """
    Real part of the dot product of 2 arrays
    """
    res = torch.sum(reprod(a, b), axis=axis)
    return res


def mad(x: torch.Tensor, dim=None):
    """
    Mean absolute deviation
    """
    return torch.mean(torch.abs(x - torch.mean(x, dim=dim)), dim=dim)



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Tuple, TYPE_CHECKING, Optional
import logging
import copy
import math

import torch
from torch import Tensor

import ptychi.image_proc as ip
import ptychi.data_structures as ds
import ptychi.data_structures.base as dsbase
import ptychi.maths as pmath
from ptychi.timing.timer_utils import timer
import ptychi.api.enums as enums
from ptychi.utils import (
    get_default_complex_dtype, 
    to_numpy, 
    chunked_processing,
    get_probe_renormalization_factor
)
import ptychi.maps as maps

if TYPE_CHECKING:
    import ptychi.api as api
    from ptychi.data_structures.probe_positions import ProbePositions
    
logger = logging.getLogger(__name__)


class SliceSpacings(dsbase.ReconstructParameter):
    options: "api.options.base.SliceSpacingOptions"

    def __init__(
        self,
        *args,
        name: str = "slice_spacings",
        options: "api.options.base.SliceSpacingOptions" = None,
        **kwargs,
    ):
        """
        Slice spacings.

        Parameters
        ----------
        data: a tensor of shape (n_slices,) giving the slice spacings in meters.
        """
        super().__init__(*args, name=name, options=options, is_complex=False, **kwargs)


class Object(dsbase.ReconstructParameter):
    options: "api.options.base.ObjectOptions"

    pixel_size_m: float = 1.0

    def __init__(
        self,
        name: str = "object",
        options: "api.options.base.ObjectOptions" = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, name=name, options=options, is_complex=True, **kwargs)
        self.pixel_size_m = options.pixel_size_m
        self.roi_bbox: dsbase.BoundingBox = None
        
        pos_origin_coords = torch.tensor(self.shape, device=torch.get_default_device()) / 2.0
        pos_origin_coords = pos_origin_coords.round() + 0.5
        self.register_buffer("pos_origin_coords", pos_origin_coords)

    def extract_patches(self, positions, patch_shape, *args, **kwargs):
        raise NotImplementedError

    def place_patches(self, positions, patches, *args, **kwargs):
        raise NotImplementedError

    @timer()
    def constrain_l1_norm(self):
        if self.options.l1_norm_constraint.weight <= 0:
            return
        data = self.data
        l1_grad = torch.sgn(data) / data.numel()
        data = data - self.options.l1_norm_constraint.weight * l1_grad
        self.set_data(data)
        logger.debug("L1 norm constraint applied to object.")
        
    def constrain_l2_norm(self):
        if self.options.l2_norm_constraint.weight <= 0:
            return
        data = self.data
        l2_grad = data * 2 / data.numel()
        data = data - self.options.l2_norm_constraint.weight * l2_grad
        self.set_data(data)
        logger.debug("L2 norm constraint applied to object.")

    def constrain_smoothness(self):
        raise NotImplementedError

    def constrain_total_variation(self) -> None:
        raise NotImplementedError

    def remove_grid_artifacts(self, *args, **kwargs):
        raise NotImplementedError
    
    def build_roi_bounding_box(self, positions: "ProbePositions"):
        pos = positions.data
        self.roi_bbox = dsbase.BoundingBox(
            sy=pos[:, 0].min(),
            ey=pos[:, 0].max(),
            sx=pos[:, 1].min(),
            ex=pos[:, 1].max(),
            origin=tuple(to_numpy(self.pos_origin_coords)),
        )
    
    def get_object_in_roi(self):
        raise NotImplementedError
    
    def update_preconditioner(self):
        raise NotImplementedError
    
    def initialize_preconditioner(self):
        raise NotImplementedError
    
    def remove_object_probe_ambiguity(self):
        raise NotImplementedError
    
    def update_pos_origin_coordinates(self, *args, **kwargs):
        raise NotImplementedError


class PlanarObject(Object):
    """
    Object that consists of one or multiple planes, i.e., 2D object or
    multislice object. The object is stored in a (n_slices, h , w) tensor.
    """

    def __init__(
        self,
        name: str = "object",
        options: "api.options.base.ObjectOptions" = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(name=name, options=options, *args, **kwargs)

        if len(self.shape) != 3:
            raise ValueError("PlanarObject should have a shape of (n_slices, h, w).")
        if (
            self.options.slice_spacings_m is not None
            and len(self.options.slice_spacings_m) != self.n_slices - 1
            and self.n_slices > 1
        ):
            raise ValueError("The number of slice spacings must be n_slices - 1.")

        # Build slice spacing object.
        if self.is_multislice:
            if self.options.slice_spacings_m is None:
                raise ValueError("slice_spacings_m must be specified for multislice objects.")
            slice_spacings_m = self.options.slice_spacings_m
        else:
            slice_spacings_m = torch.zeros(1)
            
        self.slice_spacings = SliceSpacings(
            data=slice_spacings_m,
            options=self.options.slice_spacing_options
        )
        self.register_optimizable_sub_module(self.slice_spacings)

        # Initialize position origin coordinates.
        pos_origin_coords = torch.tensor(self.shape[1:], device=torch.get_default_device()) / 2.0
        pos_origin_coords = pos_origin_coords.round() + 0.5
        self.register_buffer("pos_origin_coords", pos_origin_coords)

    @property
    def is_multislice(self) -> bool:
        return self.shape[0] > 1

    @property
    def n_slices(self):
        return self.shape[0]

    @property
    def lateral_shape(self):
        return self.shape[1:]

    @property
    def extract_patches_function(self) -> "ip.ExtractPatchesProtocol":
        return maps.get_patch_extractor_function_by_name(self.options.patch_interpolation_method)

    @property
    def place_patches_function(self) -> "ip.PlacePatchesProtocol":
        return maps.get_patch_placer_function_by_name(self.options.patch_interpolation_method)
    
    def update_pos_origin_coordinates(self, positions: Tensor = None):
        """Update the coordinates of the center pixel based on probe positions.

        Parameters
        ----------
        positions : Tensor
            A (n_pos, 2) tensor of probe positions in pixels.
        """
        if self.options.determine_position_origin_coords_by == enums.ObjectPosOriginCoordsMethods.POSITIONS:
            positions = positions.detach()
            buffer_center = torch.tensor([x / 2 for x in self.lateral_shape], device=positions.device)
            position_center = (positions.max(0).values + positions.min(0).values) / 2
            self.pos_origin_coords = buffer_center - position_center
            self.pos_origin_coords = self.pos_origin_coords.round() + 0.5
        elif self.options.determine_position_origin_coords_by == enums.ObjectPosOriginCoordsMethods.SUPPORT:
            pos_origin_coords = torch.tensor(self.shape[1:], device=torch.get_default_device()) / 2.0
            pos_origin_coords = pos_origin_coords.round() + 0.5
            self.pos_origin_coords = pos_origin_coords
        elif self.options.determine_position_origin_coords_by == enums.ObjectPosOriginCoordsMethods.SPECIFIED:
            if self.options.position_origin_coords is None:
                raise ValueError(
                    "`object_options.center_coords` should be specified when "
                    "`object_options.determine_center_coords_by` is set to "
                    "`SPECIFIED`."
                )
            self.pos_origin_coords = self.options.position_origin_coords
        else:
            raise ValueError(f"Invalid value for `determine_center_coords_by`: {self.options.determine_position_origin_coords_by}")

    def get_slice(self, index):
        return self.data[index, ...]

    @timer()
    def extract_patches(
        self, 
        positions: Tensor, 
        patch_shape: Tuple[int, int], 
        integer_mode: bool = False,
        pad_for_shift: Optional[int] = 1,
    ):
        """
        Extract (n_patches, n_slices, h', w') patches from the object.

        Parameters
        ----------
        positions : Tensor
            Tensor of shape (N, 2) giving the center positions of the patches in pixels.
            The origin of the given positions are assumed to be `self.pos_origin_coords`.
        patch_shape : tuple
            Tuple giving the lateral patch shape in pixels.
        integer_mode : bool, optional
            If True, the patches are extracted at the exact center between pixels,
            so that no interpolation is needed.
        pad_for_shift : int, optional
            If given, patches larger than the intended size by this amount are cropped
            out from the object before shifting.

        Returns
        -------
        Tensor
            Tensor of shape (N, n_slices, h', w') containing the extracted patches.
        """
        # Positions are provided with the origin in the center of the object support.
        # We shift the positions so that the origin is in the upper left corner.
        positions = positions + self.pos_origin_coords
        patches_all_slices = []
        for i_slice in range(self.n_slices):
            if integer_mode:
                patches = ip.extract_patches_integer(
                    self.get_slice(i_slice), positions, patch_shape
                )
            else:
                patches = self.extract_patches_function(
                    self.get_slice(i_slice), positions, patch_shape, pad=pad_for_shift
                )
            patches_all_slices.append(patches)
        patches_all_slices = torch.stack(patches_all_slices, dim=1)
        return patches_all_slices

    @timer()
    def place_patches(
        self, 
        positions: Tensor, 
        patches: Tensor, 
        integer_mode: bool = False,
        pad_for_shift: Optional[int] = 1,
        *args, **kwargs
    ):
        """
        Place patches into the object.

        Parameters
        ----------
        positions : Tensor
            Tensor of shape (N, 2) giving the center positions of the patches in pixels.
            The origin of the given positions are assumed to be `self.pos_origin_coords`.
        patches : Tensor
            Tensor of shape (n_patches, n_slices, H, W) of image patches.
        integer_mode : bool, optional
            If True, the patches are placed at the exact center between pixels,
            so that no interpolation is needed.
        pad_for_shift : int, optional
            If given, patches are either padded (for adjoint mode) or cropped 
            (for forward mode) before applying subpixel shifts.
        """
        positions = positions + self.pos_origin_coords
        updated_slices = []
        for i_slice in range(self.n_slices):
            if integer_mode:
                image = ip.place_patches_integer(
                    self.get_slice(i_slice), positions, patches[:, i_slice, ...], op="add"
                )
            else:
                image = self.place_patches_function(
                    self.get_slice(i_slice), positions, patches[:, i_slice, ...], op="add",
                    pad=pad_for_shift,
                )
            updated_slices.append(image)
        updated_slices = torch.stack(updated_slices, dim=0)
        self.tensor.set_data(updated_slices)

    @timer()
    def place_patches_on_empty_buffer(
        self, 
        positions: Tensor, 
        patches: Tensor, 
        integer_mode: bool = False,
        pad_for_shift: Optional[int] = 1,
        *args, **kwargs
    ):
        """
        Place patches into a zero array with the *lateral* shape of the object.

        Parameters
        ----------
        positions : Tensor
            Tensor of shape (N, 2) giving the center positions of the patches in pixels.
            The origin of the given positions are assumed to be `self.pos_origin_coords`.
        patches : Tensor
            Tensor of shape (N, H, W) of image patches.
        integer_mode : bool, optional
            If True, the patches are placed at the exact center between pixels,
            so that no interpolation is needed.
        pad_for_shift : int, optional
            If given, patches are either padded (for adjoint mode) or cropped 
            (for forward mode) before applying subpixel shifts.

        Returns
        -------
        image : Tensor
            A tensor with the lateral shape of the object with patches added onto it.
        """
        positions = positions + self.pos_origin_coords
        image = torch.zeros(
            self.lateral_shape, dtype=get_default_complex_dtype(), device=self.tensor.data.device
        )
        if integer_mode:
            image = ip.place_patches_integer(image, positions, patches, op="add")
        else:
            image = self.place_patches_function(
                image, positions, patches, op="add", pad=pad_for_shift
            )
        return image
    
    def get_object_in_roi(self):
        bbox = self.roi_bbox.get_bbox_with_top_left_origin()
        return self.data[:, int(bbox.sy):int(bbox.ey), int(bbox.sx):int(bbox.ex)]

    @timer()
    def constrain_smoothness(self) -> None:
        """
        Smooth the magnitude of the object.
        """
        if self.options.smoothness_constraint.alpha <= 0:
            return
        alpha = self.options.smoothness_constraint.alpha
        if alpha > 1.0 / 8:
            logger.warning(f"Alpha = {alpha} in smoothness constraint should be less than 1/8.")
        psf = torch.ones(3, 3, device=self.tensor.data.device) * alpha
        psf[2, 2] = 1 - 8 * alpha

        data = self.data
        for i_slice in range(self.n_slices):
            slc0 = data[i_slice]
            slc = ip.convolve2d(slc0, psf, "same")
            data[i_slice] = slc
        self.set_data(data)

    @timer()
    def constrain_total_variation(self) -> None:
        if self.options.total_variation.weight <= 0:
            return
        data = self.data
        for i_slice in range(self.n_slices):
            data[i_slice] = ip.total_variation_2d_chambolle(
                data[i_slice], lmbda=self.options.total_variation.weight, niter=2
            )
        self.set_data(data)

    @timer()
    def remove_grid_artifacts(self):
        data = self.data
        for i_slice in range(self.n_slices):
            phase = torch.angle(data[i_slice])
            phase = ip.remove_grid_artifacts(
                phase,
                pixel_size_m=self.pixel_size_m,
                period_x_m=self.options.remove_grid_artifacts.period_x_m,
                period_y_m=self.options.remove_grid_artifacts.period_y_m,
                window_size=self.options.remove_grid_artifacts.window_size,
                direction=self.options.remove_grid_artifacts.direction,
            )
            data = data[i_slice].abs() * torch.exp(1j * phase)
        self.set_data(data)

    @timer()
    def regularize_multislice(self):
        """ 
        Regularize multislice by applying a low-pass transfer function to the
        3D Fourier space of the magnitude and phase (unwrapped) of all slices.

        Adapted from fold_slice (regulation_multilayers.m).
        """
        if not self.is_multislice or self.options.multislice_regularization.weight <= 0:
            return
        if self.preconditioner is None:
            raise ValueError("Regularization requires a preconditioner.")

        # TODO: use CPU if GPU memory usage is too large.
        fourier_coords = []
        for s in (self.n_slices, self.lateral_shape[0], self.lateral_shape[1]):
            u = torch.fft.fftfreq(s)
            fourier_coords.append(u)
        fourier_coords = torch.meshgrid(*fourier_coords, indexing="ij")
        # Calculate force of regularization based on the idea that DoF = resolution^2/lambda
        w = 1 - torch.atan(
            (
                self.options.multislice_regularization.weight
                * torch.abs(fourier_coords[0])
                / torch.sqrt(fourier_coords[1] ** 2 + fourier_coords[2] ** 2 + 1e-3)
            )
            ** 2
        ) / (torch.pi / 2)
        relax = 1
        alpha = 1
        # Low-pass transfer function.
        w_a = w * torch.exp(-alpha * (fourier_coords[1] ** 2 + fourier_coords[2] ** 2))
        obj = self.data

        # Find correction for amplitude.
        aobj = torch.abs(obj)
        fobj = torch.fft.fftn(aobj)
        fobj = fobj * w_a
        aobj_upd = torch.fft.ifftn(fobj)
        # Push towards zero.
        aobj_upd = 1 + 0.9 * (aobj_upd - 1)

        # Find correction for phase.
        w_phase = torch.clip(10 * (self.preconditioner / self.preconditioner.max()), max=1)

        if self.options.multislice_regularization.unwrap_phase:
            pobj = [
                ip.unwrap_phase_2d(
                    obj[i_slice],
                    weight_map=w_phase,
                    fourier_shift_step=0.5,
                    image_grad_method=self.options.multislice_regularization.unwrap_image_grad_method,
                    image_integration_method=self.options.multislice_regularization.unwrap_image_integration_method,
                )
                for i_slice in range(self.n_slices)
            ]
            pobj = torch.stack(pobj, dim=0)
        else:
            pobj = torch.angle(obj)
        fobj = torch.fft.fftn(pobj)
        fobj = fobj * w_a
        pobj_upd = torch.fft.ifftn(fobj)

        aobj_upd = torch.real(aobj_upd) - aobj
        pobj_upd = w_phase * (torch.real(pobj_upd) - pobj)
        corr = (1 + relax * aobj_upd) * torch.exp(1j * relax * pobj_upd)
        obj = obj * corr
        self.set_data(obj)
        
    def remove_object_probe_ambiguity(
        self, 
        probe: "ds.probe.Probe", 
        *, 
        update_probe_in_place: bool = True
    ) -> Optional[Tensor]:
        """
        Remove the object-probe ambiguity by scaling the object by its norm,
        and adjusting the probe power accordingly.
        
        Parameters
        ----------
        probe : ds.probe.Probe
            The probe to adjust.
        update_probe_in_place : bool, optional
            Whether to update the probe in place. If False, the tensor of the updated probe is returned.
        """
        bbox = self.roi_bbox.get_bbox_with_top_left_origin()
        roi_slicer = bbox.get_slicer()
        w = self.preconditioner[*roi_slicer]
        w = w / pmath.mnorm(w, dim=(-2, -1))
        
        # Get the norm of the object within the ROI for each slice.
        obj_data = self.data
        obj_norm = torch.sqrt(torch.mean(torch.abs(obj_data[..., *roi_slicer]) ** 2 * w, dim=(-2, -1)))
        
        # Scale the object such that the mean transmission is 1.
        obj_data = obj_data / obj_norm[:, None, None]
        
        # Adjust the probe power accordingly.
        probe_data = probe.data
        probe_data = probe_data * torch.prod(obj_norm)
        
        self.set_data(obj_data)
        if update_probe_in_place:
            probe.set_data(probe_data)
        else:
            return probe_data
        
        
    def calculate_illumination_map(
        self, 
        probe: "ds.probe.Probe",
        probe_positions: "ds.probe_positions.ProbePositions",
        use_all_modes: bool = False
    ) -> Tensor:
        """Calculate the illumination map by overlaying the probe intensity
        at all positions on a zero array. The map is used to calculate 
        the preconditioner.
        
        Parameters
        ----------
        probe : ds.probe.Probe
            The probe to use for the illumination map.
        probe_positions : ds.probe_positions.ProbePositions
            The positions of the probe.
        use_all_modes : bool, optional
            Whether to use all modes of the probe.

        Returns
        -------
        Tensor
            The illumination map of the object.
        """
        positions_all = probe_positions.tensor
        # Shape of probe:        (n_probe_modes, h, w)
        object_ = self.get_slice(0)

        if use_all_modes:
            probe_int = probe.get_all_mode_intensity(opr_mode=0)[None, :, :]
        else:
            probe_int = probe.get_mode_and_opr_mode(mode=0, opr_mode=0)[None, ...].abs() ** 2
        # Shape of probe_int:    (n_scan_points, h, w)
        probe_int = probe_int.repeat(len(positions_all), 1, 1)

        # Stitch probes of all positions on the object buffer
        # TODO: allow setting chunk size externally
        probe_sq_map = chunked_processing(
            func=ip.place_patches_integer,
            common_kwargs={"op": "add"},
            chunkable_kwargs={
                "positions": positions_all.round().int() + self.pos_origin_coords,
                "patches": probe_int,
            },
            iterated_kwargs={
                "image": torch.zeros_like(object_.real).type(torch.get_default_dtype())
            },
            chunk_size=64,
        )
        return probe_sq_map

    def update_preconditioner(
        self,
        probe: "ds.probe.Probe",
        probe_positions: "ds.probe_positions.ProbePositions",
        patterns: Tensor = None,
        use_all_modes: bool = False,
    ) -> None:
        """Update the preconditioner. This function reproduces the behavior
        of PtychoShelves in `ptycho_solver`: it averages the new illumination
        map and the old preconditioner to reduce oscillations.
        
        Parameters
        ----------
        probe : ds.probe.Probe
            The probe to use for the illumination map.
        probe_positions : ds.probe_positions.ProbePositions
            The positions of the probe.
        patterns : Tensor, optional
            A (n_scan_points, h, w) tensor giving the diffraction patterns. Only needed
            if the preconditioner does not exist and needs to be initialized.
        use_all_modes : bool, optional
            Whether to use the sum of all probe modes' intensities for the illumination map.
        """
        if self.preconditioner is None:
            self.initialize_preconditioner(probe, probe_positions, patterns)
        illum_map = self.calculate_illumination_map(probe, probe_positions, use_all_modes=use_all_modes)
        self.preconditioner = (self.preconditioner + illum_map) / 2
    
    def initialize_preconditioner(
        self, 
        probe: "ds.probe.Probe", 
        probe_positions: "ds.probe_positions.ProbePositions",
        patterns: Tensor,
    ) -> None:
        """Initialize the preconditioner. This function reproduces the behavior
        of PtychoShelves in `init_solver` and `load_from_p`: the probe is first
        renormalized before being used to calculate the illumination map.
        Diffraction patterns are needed to calculate the renormalization factor.
        
        Parameters
        ----------
        probe : ds.probe.Probe
            The probe to use for the illumination map.
        probe_positions : ds.probe_positions.ProbePositions
            The positions of the probe.
        patterns : ds.patterns.Patterns
            A (n_scan_points, h, w) tensor giving the diffraction patterns.
        """
        probe_data = probe.data
        probe_renormalization_factor = get_probe_renormalization_factor(patterns)
        probe_data = probe_data / (math.sqrt(probe.shape[-1] * probe.shape[-2]) * 2 * probe_renormalization_factor)
        probe_temp = ds.probe.Probe(data=probe_data, options=copy.deepcopy(probe.options))
        self.preconditioner = self.calculate_illumination_map(probe_temp, probe_positions, use_all_modes=False)


class DIPObject(Object):
    
    options: "api.options.ad_ptychography.AutodiffPtychographyObjectOptions"
    
    def __init__(
        self,
        name: str = "object",
        options: "api.options.ad_ptychography.AutodiffPtychographyObjectOptions" = None,
        *args,
        **kwargs
    ) -> None:
        """Deep image prior object.

        Parameters
        ----------
        name : str, optional
            The name of the object.
        options : api.options.base.ObjectOptions, optional
            The options for the object.
        """
        super().__init__(name=name, options=options, *args, **kwargs)
        self.model = None
        self.dip_output_magnitude = None
        self.dip_output_phase = None
        
        self.build_model()
        self.build_dip_optimizer()
        
        # `self.tensor` is used to hold the object generated by the DIP model and
        # is not trainable.
        self.tensor.requires_grad_(False)
        
        self.initial_data = None
        if self.options.experimental.deep_image_prior_options.residual_generation:
            self.initial_data = self.data.clone()   
        
    def build_model(self):
        if not self.options.experimental.deep_image_prior_options.enabled:
            return
        model_class = maps.get_nn_model_by_enum(self.options.experimental.deep_image_prior_options.model)
        self.model = model_class(
            **self.options.experimental.deep_image_prior_options.model_params
        )
        
    def build_dip_optimizer(self):
        if self.optimizable and self.optimizer_class is None:
            raise ValueError(
                "Parameter {} is optimizable but no optimizer is specified.".format(self.name)
            )
        if self.optimizable:
            self.optimizer = self.optimizer_class(self.model.parameters(), **self.optimizer_params)
        
    def generate(self):
        raise NotImplementedError


class DIPPlanarObject(DIPObject, PlanarObject):
    def __init__(
        self,
        name: str = "object",
        options: "api.options.base.ObjectOptions" = None,
        *args,
        **kwargs
    ) -> None:
        """Deep image prior planar object.

        Parameters
        ----------
        name : str, optional
            The name of the object.
        options : api.options.base.ObjectOptions, optional
            The options for the object.
        """
        super().__init__(name=name, options=options, data_as_parameter=False, *args, **kwargs)
        
        nn_input = self.get_nn_input()
        self.register_buffer("nn_input", nn_input)
        
    def get_nn_input(self):
        z = torch.rand(
            [self.n_slices, self.options.experimental.deep_image_prior_options.net_input_channels, *self.lateral_shape], 
        ) * 0.1
        return z

    def generate(self) -> Tensor:
        """Generate the object using the deep image prior model, and set the
        generated object to self.data.
        
        Returns
        -------
        Tensor
            The generated object.
        """
        if self.model is None:
            raise ValueError("Model is not built.")
        o = self.model(self.nn_input)
        
        o, mag, phase = self.process_net_output(o)
        
        with torch.no_grad():
            self.dip_output_magnitude = mag.clone()
            self.dip_output_phase = phase.clone()
        
        if self.options.experimental.deep_image_prior_options.residual_generation:
            init_data = torch.stack([self.initial_data.real, self.initial_data.imag], dim=-1)
            o = o + init_data
        self.tensor.data = o
        return self.data

    def process_net_output(self, o):
        """Apply constraints to the DIP network's output and transform the tensor.
        For the magnitude, it is clipped at 0 and then added by 1.0.

        Parameters
        ----------
        o : Tensor | tuple[Tensor, Tensor]
            The output of the DIP network. It should either be a [n_slices, 2, h, w] 
            tensor with the channels giving the magnitude and phase of the object,
            or a tuple of two [n_slices, 1, h, w] tensors giving the magnitude and phase
            of the object.
            
        Returns
        -------
        Tensor
            A [n_slices, h, w, 2] tensor representing the real and imaginary parts 
            of the object.
        Tensor
            The magnitude of the object.
        Tensor
            The phase of the object.
        """
        if isinstance(o, (list, tuple)):
            mag, phase = o
            mag = mag[:, 0]
            phase = phase[:, 0]
        else:
            mag = o[:, 0]
            phase = o[:, 1]
            
        expected_phase_shape = (self.n_slices, *self.lateral_shape)
        if tuple(phase.shape) != expected_phase_shape:
            logger.warning(
                "Shape of phase output by NN is expected to be "
                f"(n_slices, h, w) = {expected_phase_shape} but got {phase.shape}. "
                "Padding/cropping generated object to match expected shape."
            )
            mag = ip.central_crop_or_pad(mag, expected_phase_shape[1:])
            phase = ip.central_crop_or_pad(phase, expected_phase_shape[1:])
        
        if self.options.experimental.deep_image_prior_options.constrain_object_outside_network:
            mag = torch.sigmoid(mag)
        
        o_complex = mag * torch.exp(1j * phase)
        o = torch.stack([o_complex.real, o_complex.imag], dim=-1)
        return o, mag, phase



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Type

import torch

import ptychi.api.enums as enums
from ptychi.metrics import MSELossOfSqrt
import ptychi.reconstructors as reconstructors
import ptychi.forward_models as fm
from ptychi.reconstructors.base import Reconstructor
import ptychi.image_proc as ip
import ptychi.reconstructors.nn as pnn
from functools import partial


def get_complex_dtype_by_enum(key: enums.Dtypes) -> torch.dtype:
    return {enums.Dtypes.FLOAT32: torch.complex64, enums.Dtypes.FLOAT64: torch.complex128}[key]


def get_optimizer_by_enum(key: enums.Optimizers) -> torch.optim.Optimizer:
    return {
        enums.Optimizers.SGD: torch.optim.SGD,
        enums.Optimizers.ADAM: torch.optim.Adam,
        enums.Optimizers.RMSPROP: torch.optim.RMSprop,
        enums.Optimizers.ADAGRAD: torch.optim.Adagrad,
        enums.Optimizers.ADADELTA: torch.optim.Adadelta,
        enums.Optimizers.LBFGS: torch.optim.LBFGS,
        enums.Optimizers.ASGD: torch.optim.ASGD,
        enums.Optimizers.SPARSE_ADAM: torch.optim.SparseAdam,
        enums.Optimizers.ADAMAX: torch.optim.Adamax,
        enums.Optimizers.RADAM: torch.optim.RAdam,
        enums.Optimizers.ADAMW: torch.optim.AdamW,
    }[key]


def get_loss_function_by_enum(key: enums.LossFunctions) -> torch.nn.Module:
    return {
        enums.LossFunctions.MSE: torch.nn.MSELoss,
        enums.LossFunctions.POISSON: torch.nn.PoissonNLLLoss,
        enums.LossFunctions.MSE_SQRT: MSELossOfSqrt,
    }[key]


def get_forward_model_by_enum(key: enums.ForwardModels) -> Type["fm.ForwardModel"]:
    return {
        enums.ForwardModels.PLANAR_PTYCHOGRAPHY: fm.PlanarPtychographyForwardModel,
    }[key]


def get_reconstructor_by_enum(key: enums.Reconstructors) -> Type["Reconstructor"]:
    return {
        enums.Reconstructors.base: Reconstructor,
        enums.Reconstructors.LSQML: reconstructors.LSQMLReconstructor,
        enums.Reconstructors.AD_PTYCHO: reconstructors.AutodiffPtychographyReconstructor,
        enums.Reconstructors.PIE: reconstructors.PIEReconstructor,
        enums.Reconstructors.EPIE: reconstructors.EPIEReconstructor,
        enums.Reconstructors.RPIE: reconstructors.RPIEReconstructor,
        enums.Reconstructors.DM: reconstructors.DMReconstructor,
        enums.Reconstructors.BH: reconstructors.BHReconstructor,
    }[key]


def get_noise_model_by_enum(key: enums.NoiseModels) -> str:
    return {enums.NoiseModels.GAUSSIAN: "gaussian", enums.NoiseModels.POISSON: "poisson"}[key]


def get_device_by_enum(key: enums.Devices) -> str:
    return {enums.Devices.CPU: "cpu", enums.Devices.GPU: "cuda"}[key]


def get_dtype_by_enum(key: enums.Dtypes) -> torch.dtype:
    return {enums.Dtypes.FLOAT32: torch.float32, enums.Dtypes.FLOAT64: torch.float64}[key]


def get_patch_placer_function_by_name(
    key: enums.PatchInterpolationMethods,
) -> "ip.PlacePatchesProtocol":
    return {
        enums.PatchInterpolationMethods.FOURIER: ip.place_patches_fourier_shift,
        enums.PatchInterpolationMethods.BILINEAR: partial(
            ip.place_patches_bilinear_shift, round_positions=False
        ),
        enums.PatchInterpolationMethods.NEAREST: partial(
            ip.place_patches_bilinear_shift, round_positions=True
        ),
    }[key]


def get_patch_extractor_function_by_name(
    key: enums.PatchInterpolationMethods,
) -> "ip.ExtractPatchesProtocol":
    return {
        enums.PatchInterpolationMethods.FOURIER: ip.extract_patches_fourier_shift,
        enums.PatchInterpolationMethods.BILINEAR: partial(
            ip.extract_patches_bilinear_shift, round_positions=False
        ),
        enums.PatchInterpolationMethods.NEAREST: partial(
            ip.extract_patches_bilinear_shift, round_positions=True
        ),
    }[key]


def get_nn_model_by_enum(key: enums.DIPModels) -> Type["torch.nn.Module"]:
    return {
        enums.DIPModels.UNET: pnn.models.unet.UNet,
        enums.DIPModels.AUTOENCODER: pnn.models.autoencoder.Autoencoder,
    }[key]



import argparse

import torch

import ptychi.api as api
from ptychi.api.task import PtychographyTask
from ptychi.utils import get_suggested_object_size, get_default_complex_dtype

import test_utils as tutils


class Test2DEPIEPositionCorrection(tutils.TungstenDataTester):

    @tutils.BaseTester.wrap_recon_tester(name='test_2d_ptycho_epie_position_correction')
    def test_2d_ptycho_epie_position_correction(self):        
        self.setup_ptychi(cpu_only=False)
        
        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0)
        probe = probe[:, [0], :, :]
        
        options = api.EPIEOptions()
        
        options.data_options.data = data
        
        options.object_options.initial_guess = torch.ones([1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], dtype=get_default_complex_dtype())
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True
        options.object_options.optimizer = api.Optimizers.SGD
        options.object_options.step_size = 0.1
        options.object_options.alpha = 1
        
        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.optimizer = api.Optimizers.SGD
        options.probe_options.step_size = 0.1
        options.probe_options.alpha = 1

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = True
        options.probe_position_options.optimizer = api.Optimizers.SGD
        options.probe_position_options.step_size = 1000
        options.probe_position_options.correction_options.correction_type = api.PositionCorrectionTypes.CROSS_CORRELATION
        
        options.reconstructor_options.batch_size = 96
        options.reconstructor_options.num_epochs = 32
        options.reconstructor_options.allow_nondeterministic_algorithms = False
        options.reconstructor_options.use_double_precision_for_fft = True
        
        task = PtychographyTask(options)
        task.run()

        recon = task.get_data_to_cpu('object', as_numpy=True)[0]
        return recon
    
    
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--generate-gold', action='store_true')
    args = parser.parse_args()

    tester = Test2DEPIEPositionCorrection()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_2d_ptycho_epie_position_correction()



import argparse

import torch

import ptychi.api as api
from ptychi.api.task import PtychographyTask
from ptychi.utils import get_suggested_object_size, get_default_complex_dtype

import test_utils as tutils


class Test2DPtychoBH(tutils.TungstenDataTester):

    @tutils.TungstenDataTester.wrap_recon_tester(name='test_2d_ptycho_bh_obj')
    def test_2d_ptycho_bh_obj(self):
        self.setup_ptychi(cpu_only=False)

        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0)
        probe = probe[:, [0], :, :]

        options = api.BHOptions()
        options.data_options.data = data

        options.object_options.initial_guess = torch.ones([1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], dtype=get_default_complex_dtype())
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True

        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = False
        options.probe_options.rho = 0.1

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = False
        options.probe_position_options.rho = 2

        options.reconstructor_options.method = 'GD'
        options.reconstructor_options.batch_size = 96
        options.reconstructor_options.num_epochs = 16
        options.reconstructor_options.allow_nondeterministic_algorithms = False
        options.reconstructor_options.use_double_precision_for_fft = False
        options.reconstructor_options.forward_model_options.pad_for_shift = 4

        task = PtychographyTask(options)
        task.run()

        recon = task.get_data_to_cpu('object', as_numpy=True)[0]

        return recon
    
    @tutils.TungstenDataTester.wrap_recon_tester(name='test_2d_ptycho_bh_obj_probe')
    def test_2d_ptycho_bh_obj_probe(self):
        self.setup_ptychi(cpu_only=False)

        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0)
        probe = probe[:, [0], :, :]

        options = api.BHOptions()
        options.data_options.data = data

        options.object_options.initial_guess = torch.ones([1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], dtype=get_default_complex_dtype())
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True

        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.rho = 0.1

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = False
        options.probe_position_options.rho = 2

        options.reconstructor_options.method = 'GD'
        options.reconstructor_options.batch_size = 96
        options.reconstructor_options.num_epochs = 16
        options.reconstructor_options.allow_nondeterministic_algorithms = False
        options.reconstructor_options.use_double_precision_for_fft = False
        options.reconstructor_options.forward_model_options.pad_for_shift = 4

        task = PtychographyTask(options)
        task.run()

        recon = task.get_data_to_cpu('object', as_numpy=True)[0]

        return recon 
    
    @tutils.TungstenDataTester.wrap_recon_tester(name='test_2d_ptycho_bh_obj_probe_pos')
    def test_2d_ptycho_bh_obj_probe_pos(self):
        self.setup_ptychi(cpu_only=False)

        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0)
        probe = probe[:, [0], :, :]

        options = api.BHOptions()
        options.data_options.data = data

        options.object_options.initial_guess = torch.ones([1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], dtype=get_default_complex_dtype())
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True

        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.rho = 0.1

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = True
        options.probe_position_options.rho = 2

        options.reconstructor_options.method = 'GD'
        options.reconstructor_options.batch_size = 96
        options.reconstructor_options.num_epochs = 16
        options.reconstructor_options.allow_nondeterministic_algorithms = False
        options.reconstructor_options.use_double_precision_for_fft = False
        options.reconstructor_options.forward_model_options.pad_for_shift = 4

        task = PtychographyTask(options)
        task.run()

        recon = task.get_data_to_cpu('object', as_numpy=True)[0]

        return recon 
    
    @tutils.TungstenDataTester.wrap_recon_tester(name='test_2d_ptycho_bh_obj_probe_pos_cg')
    def test_2d_ptycho_bh_obj_probe_pos_cg(self):
        self.setup_ptychi(cpu_only=False)

        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0)
        probe = probe[:, [0], :, :]

        options = api.BHOptions()
        options.data_options.data = data

        options.object_options.initial_guess = torch.ones([1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], dtype=get_default_complex_dtype())
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True

        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.rho = 0.1

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = True
        options.probe_position_options.rho = 2

        options.reconstructor_options.method = 'CG'
        options.reconstructor_options.batch_size = 96
        options.reconstructor_options.num_epochs = 8
        options.reconstructor_options.allow_nondeterministic_algorithms = False
        options.reconstructor_options.use_double_precision_for_fft = False
        options.reconstructor_options.forward_model_options.pad_for_shift = 4

        task = PtychographyTask(options)
        task.run()

        recon = task.get_data_to_cpu('object', as_numpy=True)[0]

        return recon 


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--generate-gold', action='store_true')
    args = parser.parse_args()

    tester = Test2DPtychoBH()
    tester.setup_method(name="", generate_data=False,
                        generate_gold=args.generate_gold, debug=True)
    tester.test_2d_ptycho_bh_obj()
    tester.test_2d_ptycho_bh_obj_probe()
    tester.test_2d_ptycho_bh_obj_probe_pos()
    tester.test_2d_ptycho_bh_obj_probe_pos_cg()
    


# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional, Union
import logging
import math

import torch
import torch.utils
from torch.utils.data import Dataset
from torch import Tensor
from numpy import ndarray
import sklearn.cluster
import numpy as np
import scipy.spatial

from ptychi.utils import to_tensor, to_numpy
import ptychi.maths as pmath

logger = logging.getLogger(__name__)


class PtychographyDataset(Dataset):
    def __init__(
        self,
        patterns: Union[Tensor, ndarray],
        valid_pixel_mask: Optional[Union[Tensor, ndarray]] = None,
        wavelength_m: float = None,
        free_space_propagation_distance_m: float = 1.0,
        fft_shift: bool = True,
        save_data_on_device: bool = False,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.patterns = to_tensor(patterns, device="cpu" if not save_data_on_device else "cuda")
        if fft_shift:
            self.patterns = torch.fft.fftshift(self.patterns, dim=(-2, -1))
            logger.info("Diffraction data have been FFT-shifted.")

        if valid_pixel_mask is None:
            valid_pixel_mask = torch.ones(self.patterns.shape[-2:])
        self.valid_pixel_mask = to_tensor(valid_pixel_mask, device="cpu", dtype=torch.bool)
        
        if fft_shift:
            self.valid_pixel_mask = torch.fft.fftshift(self.valid_pixel_mask, dim=(-2, -1))
            logger.info("Valid pixel mask have been FFT-shifted.")

        self.wavelength_m = wavelength_m
        self.free_space_propagation_distance_m = free_space_propagation_distance_m
        
        self.save_data_on_device = save_data_on_device

    def __getitem__(self, index):
        if not isinstance(index, torch.Tensor):
            index = torch.tensor(index, device=self.patterns.device, dtype=torch.long)
        pattern = self.patterns[index]
        return index, pattern

    def __len__(self):
        return len(self.patterns)

    def move_attributes_to_device(self, device=None):
        if device is None:
            device = torch.get_default_device()
        self.valid_pixel_mask = self.valid_pixel_mask.to(device)


class PtychographyCompactBatchSampler(torch.utils.data.Sampler):
    def __init__(self, positions, batch_size, *args, **kwargs):
        """
        A batch sampler that returns minibatches containing indices
        that are clustered together using k-Means. This is equivalent
        to the "compact" mode in PtychoShelves. This implementation is
        adapted from `get_close_indices.m` in PtychoShelves.

        Parameters
        ----------
        positions : Tensor
            A (N, 2) tensor of probe positions in pixels.
        batch_size : int
            The batch size.
        """
        self.positions = to_numpy(positions)
        self.batch_size = batch_size
        self.clusters_of_indices = []
        
        self.build_clusters()
        
    def __len__(self):
        if len(self.clusters_of_indices) == 0:
            raise ValueError("No clusters have been built yet. Call `build_clusters()` first.")
        return len(self.clusters_of_indices)

    def __iter__(self):
        for i_batch in range(len(self)):
            yield self.clusters_of_indices[i_batch]
            
    def estimate_n_clusters_iniitially(self):
        return int(np.ceil(len(self.positions) / self.batch_size))

    def build_clusters(self):
        self.clusters_of_indices = []
        cluster_labels, centers, _, distmat = self.get_best_kmeans()
        cluster_labels, centers, distmat = self.refine_clusters(cluster_labels, centers, distmat)
        
        # Sort cluster labels, so that the largest cluster is labeled as 0.
        n_clusters = len(centers)
        bin_populations, _ = np.histogram(cluster_labels, bins=n_clusters)
        sorted_inds = np.argsort(bin_populations)[::-1]
        
        # Populate batch indices.
        for i_cluster in range(n_clusters):
            self.clusters_of_indices.append(
                to_tensor(
                    np.where(cluster_labels == sorted_inds[i_cluster])[0], 
                    device="cpu", dtype=torch.long)
            )
        
        # PtychoShelves would truncate the clusters to the smallest batch size and throw
        # away some diffraction patterns if the range of cluster population is equal 
        # to or less than 1 to make it run faster on GPU. We don't do it here.
        return

    def update_clusters(self, positions):
        self.positions = to_numpy(positions)
        self.build_clusters()
        
    def get_best_kmeans(self, n_trials=10):
        """Run multiple k-Means trials and return the best one that has the
        most uniform cluster populations.
        
        Parameters
        ----------
        n_trials : int
            The number of k-Means trials to run.
            
        Returns
        -------
        ndarray
            An (n_positions,) array of cluster labels of the best run.
        ndarray
            An (n_clusters, 2) array of cluster centers of the best run.
        ndarray
            An (n_clusters, ) array giving the sum of squared distances of all
            points belonging to a cluster to the cluster center, for all
            clusters of the best run. 
        ndarray
            An (n_positions, n_clusters) matrix giving the squared distance of each
            point to each cluster center of the best run.
        """
        best_score = np.inf
        best_cluster_labels = None
        best_centers = None
        best_sumd = np.inf
        best_distmat = None
        for _ in range(n_trials):
            kmeans_engine = sklearn.cluster.KMeans(n_clusters=self.estimate_n_clusters_iniitially())
            kmeans_engine.fit(self.positions)
            cluster_labels = kmeans_engine.predict(self.positions)
            bin_populations = np.histogram(cluster_labels, bins=kmeans_engine.n_clusters)[0]
            score = np.std(bin_populations)
            if score < best_score:
                best_score = score
                best_cluster_labels = cluster_labels
                best_centers = kmeans_engine.cluster_centers_
                best_distmat = scipy.spatial.distance.cdist(self.positions, best_centers) ** 2
                best_sumd = self.get_within_cluster_distances(self.positions, best_cluster_labels, best_centers)
        return best_cluster_labels, best_centers, best_sumd, best_distmat
    
    def refine_clusters(self, cluster_labels, centers, distmat):
        iter = 0
        n_clusters = len(centers)
        n_pos = len(self.positions)
        
        adjusted_batch_size = int(np.ceil(n_pos / np.ceil(n_pos / self.batch_size)))
        
        # Iteratively homogenize the cluster populations by moving points belonging
        # to large clusters and are closest to the smallest cluster's center to the 
        # smallest cluster.
        while True:
            iter += 1
            bin_populations, _ = np.histogram(cluster_labels, bins=n_clusters)
            n_unique_populations = len(np.unique(bin_populations))
            if n_unique_populations == 1:
                break
            if n_unique_populations < max(2, math.ceil(iter / 1e3)) and (
                n_clusters * self.batch_size != n_pos or iter > 1e3
            ):
                break
            
            # Find the cluster with the smallest population and add new points to it.
            ind_smallest_cluster = np.argmin(bin_populations)
            inds_large_cluster = np.where(bin_populations > adjusted_batch_size)[0]
            if len(inds_large_cluster) == 0 or ind_smallest_cluster in inds_large_cluster:
                break
            # Find the index (indices) of the point(s) that
            # (1) belong to large clusters, and
            # (2) are closest to the smallest cluster.
            # We can have multiple such points if there is a tie in the distance.
            mask_points_in_large_clusters = np.isin(cluster_labels, inds_large_cluster)
            mask_points_closest_to_smallest_cluster = \
                distmat[:, ind_smallest_cluster] == np.min(
                    distmat[mask_points_in_large_clusters, ind_smallest_cluster]
                )
            # Assign these points to the smallest cluster.
            cluster_labels[mask_points_closest_to_smallest_cluster] = ind_smallest_cluster
            
        # Remove empty clusters. Labels of the new clusters are still consecutive.
        #         |<- points ->|
        #        | 0 1 0 0                                 0 0 0 0 
        # clusters 1 0 1 0 ...   =replace 1s with labels=> 1 0 1 0 ... =sum=> 1 0 1 2 ...
        #        | 0 0 0 1                                 0 0 0 2
        unique_cluster_labels = np.unique(cluster_labels)
        n_clusters = len(unique_cluster_labels)
        cluster_labels = np.sum(
            np.arange(n_clusters)[:, None] * (cluster_labels == unique_cluster_labels[:, None]), 
            axis=0
        )
        cluster_labels = cluster_labels.astype(int)
        
        # Rebuild cluster centers and distance matrix.
        cluster_centers = np.zeros((n_clusters, 2))
        for i_cluster in range(n_clusters):
            cluster_centers[i_cluster, :] = np.median(self.positions[cluster_labels == i_cluster], axis=0)
        distmat = scipy.spatial.distance.cdist(self.positions, cluster_centers) ** 2
        
        # Find more compact refinement.
        # For each point, get a list of cluster indices ranked by the distance 
        # between the point and the cluster centers.
        sind = np.argsort(distmat, axis=1)
        optimal_cluster_labels = sind[:, 0]
        
        # Iteratively find the points that are non-optimal (i.e., belong to a cluster whose
        # center is not the closest one) and swap them to the optimal cluster.
        largest_nonoptimal_ratio = 1
        for iter in range(10):
            mask_nonoptimal_points = cluster_labels != optimal_cluster_labels
            nonoptimal_ratio = np.mean(mask_nonoptimal_points)
            if nonoptimal_ratio > largest_nonoptimal_ratio:
                break
            largest_nonoptimal_ratio = nonoptimal_ratio
            
            for i in range(np.sum(mask_nonoptimal_points)):
                dists_to_cluster_centers = distmat[tuple(range(n_pos)), cluster_labels]
                ind_worst = pmath.masked_argmax(dists_to_cluster_centers, mask_nonoptimal_points)
                cluster_old = cluster_labels[ind_worst]
                cluster_new = optimal_cluster_labels[ind_worst]
                # Find the index of the point in cluster_new to swap with. This is the point
                # that is closest to the center of cluster_old.
                ind_new = pmath.masked_argmin(distmat[:, cluster_old], cluster_labels == cluster_new)
                
                cluster_labels[ind_worst] = cluster_new
                cluster_labels[ind_new] = cluster_old
        
                mask_nonoptimal_points[np.array([ind_worst, ind_new])] = 0
                if np.all(mask_nonoptimal_points == 0):
                    break
        return cluster_labels, cluster_centers, distmat

    @staticmethod
    def get_within_cluster_distances(positions, cluster_labels, centers):
        """Calculate the sum of squared distances of all points belonging to a cluster
        to the cluster center.
        
        Parameters
        ----------
        positions : ndarray
            A (n_positions, 2) array of probe positions in pixels.
        cluster_labels : ndarray
            An (n_positions,) array of cluster labels.
        centers : ndarray
            An (n_clusters, 2) array of cluster centers.
        """
        n_clusters = len(centers)
        sumds = np.zeros(n_clusters)
        for i_cluster in range(n_clusters):
            inds_of_cluster = np.where(cluster_labels == i_cluster)[0]
            sq_dists = np.sum((positions[inds_of_cluster] - centers[i_cluster]) ** 2, axis=1)
            sumds[i_cluster] = np.sum(sq_dists)
        return sumds


class PtychographyUniformBatchSampler(torch.utils.data.Sampler):
    def __init__(self, positions, batch_size, *args, **kwargs):
        """A batch sampler that returns minibatches of indices that are random,
        but modified to be spread out over the space as uniformly as possible.
        This sampler is equivalent to the "sparse" mode in PtychoShelves.

        Parameters
        ----------
        positions : Tensor
            A (N, 2) tensor of probe positions in pixels.
        batch_size : int
            The batch size.
        """
        self.positions = positions
        self.batch_size = batch_size
        
        self.build_indices()

    def __len__(self):
        return math.ceil(len(self.positions) / self.batch_size)

    def __iter__(self):
        for i in np.random.choice(range(len(self)), len(self), replace=False):
            yield self.batches_of_indices[i]

    def build_indices(self):
        dist_mat = torch.cdist(self.positions, self.positions, p=2)
        dist_mat[dist_mat == 0] = torch.inf
        batches_of_indices = torch.split(
            torch.randperm(len(self.positions), device=dist_mat.device), self.batch_size
        )

        for i_batch in range(len(self) - 1):
            inds_current_batch = batches_of_indices[i_batch]
            dist_mat_current_batch = dist_mat[inds_current_batch, :][:, inds_current_batch]
            for i_pass in range(2):
                for ind_next_batch in range(len(batches_of_indices[i_batch + 1])):
                    # Calculate the inverse-distance weighted average of the distances to other
                    # points in the batch for every point:
                    # $s_i = \frac{1}{n} sum_j(d_{ij} * 1/d_{ij}) / sum_j(1/d_{ij}) = 1 / sum_j(1 / d_{ij})$
                    # Closer neighbors are more heavily
                    # weighted. This prevents the algorithm from always selecting the point
                    # around the center of the batch as in the case of a simple sum.
                    avgs_of_dists = 1.0 / torch.sum(1.0 / dist_mat_current_batch**2, dim=0)
                    if torch.all(torch.isinf(avgs_of_dists)):
                        break

                    # Swap the scan point with the smallest summed distance to other points in the batch
                    # with the scan point in the next batch.
                    min_dist_ind = torch.min(avgs_of_dists, dim=0).indices
                    (
                        batches_of_indices[i_batch][min_dist_ind],
                        batches_of_indices[i_batch + 1][ind_next_batch],
                    ) = (
                        batches_of_indices[i_batch + 1][ind_next_batch].item(),
                        batches_of_indices[i_batch][min_dist_ind].item(),
                    )

                    # Update distance matrix.
                    dist_mat_current_batch_update = dist_mat[
                        batches_of_indices[i_batch][min_dist_ind], batches_of_indices[i_batch]
                    ].clone()
                    dist_mat_current_batch[min_dist_ind, :] = dist_mat_current_batch_update
                    dist_mat_current_batch[:, min_dist_ind] = dist_mat_current_batch_update
        self.batches_of_indices = batches_of_indices



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

import logging
import dataclasses
from typing import Optional, TYPE_CHECKING
import torch
from torch import Tensor
from torch.nn import ModuleList

import ptychi.data_structures.object
import ptychi.data_structures.probe
from ptychi.propagate import (
    WavefieldPropagatorParameters,
    AngularSpectrumPropagator,
    FourierPropagator,
    FresnelTransformPropagator
)
from ptychi.metrics import MSELossOfSqrt
import ptychi.image_proc as ip
from ptychi.timing.timer_utils import timer

if TYPE_CHECKING:
    import ptychi.data_structures.parameter_group
    import ptychi.data_structures.base as dsbase
    
logger = logging.getLogger(__name__)


@dataclasses.dataclass
class IntermediateVariables(dict):
    def __getitem__(self, key):
        return getattr(self, key)
    
    def __setitem__(self, key, value):
        return setattr(self, key, value)


class ForwardModel(torch.nn.Module):
    def __init__(
        self,
        parameter_group: "ptychi.data_structures.parameter_group.ParameterGroup",
        retain_intermediates: bool = False,
        detector_size: Optional[tuple[int, int]] = None,
        *args,
        **kwargs,
    ) -> None:
        """
        The base forward model class.
        
        Parameters
        ----------
        parameter_group : ptychi.data_structures.parameter_group.ParameterGroup
            The parameter group.
        retain_intermediates : bool
            If True, intermediate variables are retained in `self.intermediate_variables`.
        detector_size : tuple[int, int], optional
            The size of the detector. If given and if the detector size is smaller than
            the size of the probe, the probe is cropped to the detector size.
        """
        super().__init__()

        self.parameter_group = parameter_group
        self.retain_intermediates = retain_intermediates
        self.optimizable_parameters: ModuleList[dsbase.ReconstructParameter] = ModuleList()
        self.propagator = None
        self.detector_size = detector_size
        self.intermediate_variables = {}
        
        self.register_optimizable_parameters()

    def register_optimizable_parameters(self):
        for var in self.parameter_group.__dict__.values():
            if var.optimizable:
                self.optimizable_parameters.append(var)

    def forward(self, *args, **kwargs):
        raise NotImplementedError

    def post_differentiation_hook(self, *args, **kwargs):
        pass

    def record_intermediate_variable(self, name, var):
        if self.retain_intermediates:
            if isinstance(self.intermediate_variables[name], list):
                self.intermediate_variables[name].append(var.detach())
            else:
                self.intermediate_variables[name] = var.detach()


class PlanarPtychographyForwardModel(ForwardModel):
    @dataclasses.dataclass
    class PlanarPtychographyIntermediateVariables(IntermediateVariables):
        positions: Tensor = None
        """Tensor of shape `[batch_size, 2]`, probe positions in the 
        current batch.
        """
        
        obj_patches: Tensor = None
        """Tensor of shape `[batch_size, n_slices, h, w]`, object patches of the 
        current batch.
        """
        
        psi: Tensor = None
        """Tensor of shape `[batch_size, n_probe_modes, h, w]`, wavefields at 
        the exit plane. For multislice objects, this is the wavefield at the 
        exit plane of the last slice.
        """
        
        psi_far: Tensor = None
        """Tensor of shape `[batch_size, n_probe_modes, h, w]`, wavefields at 
        the far field.
        """
        
        shifted_unique_probes: list[Tensor] = dataclasses.field(default_factory=list)
        """List of `n_slices` tensors of shape `[batch_size, n_probe_modes, h, w]`, unique 
        probes at each slice for every position. The position-specificity is due 
        to OPR modes and/or subpixel shifts 
        (when `apply_subpixel_shifts_on_probe == True`).
        """
    
    def __init__(
        self,
        parameter_group: "ptychi.data_structures.parameter_group.PlanarPtychographyParameterGroup",
        retain_intermediates: bool = False,
        detector_size: Optional[tuple[int, int]] = None,
        wavelength_m: Optional[float] = None,
        free_space_propagation_distance_m: Optional[float] = torch.inf,
        pad_for_shift: Optional[int] = 0,
        apply_subpixel_shifts_on_probe: bool = True,
        low_memory_mode: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        The forward model for planar ptychography, which supports multislice propagation
        but does not work for dense 3D objects that requires rotation. 
        
        While running the forward model, some intermediate variables are recorded in 
        `self.intermediate_variables` to be used by the reconstructor. For more details,
        see `PlanarPtychographyForwardModel.PlanarPtychographyIntermediateVariables`.

        Parameters
        ----------
        parameter_group : ptychi.data_structures.parameter_group.PlanarPtychographyParameterGroup
            The parameter group.
        retain_intermediates : bool
            If True, intermediate variables are retained in `self.intermediate_variables`.
        detector_size : tuple[int, int], optional
            The size of the detector. If given and if the detector size is smaller than
            the size of the probe, the probe is cropped to the detector size.
        wavelength_m : float, optional
            The wavelength of the probe in meters.
        free_space_propagation_distance_m : float, optional
            The free-space propagation distance in meters, or `inf` for far-field.
        pad_for_shift : int, optional
            If given, the probe (or object when `apply_subpixel_shifts_on_probe == False`) 
            is padded with border values by this amount before shifting.
        apply_subpixel_shifts_on_probe : bool
            If True, the subpixel parts of probe positions are accounted for by uniquely shifting
            the probe for each position, while all object patches are extracted at integer positions.
            This creates a unique probe for each position, stored in 
            `self.intermediate_variables["shifted_unique_probes"]`. If False, subpixel sihfts
            are applied on the object patches, and a common probe is used for all positions.
            The choice of the subpixel handling scheme must be consistent with the reconstructor.
        low_memory_mode : bool
            If True, incoherent probe modes are propagated sequentially rather than in parallel
            which reduces the memory usage.
        """
        super().__init__(
            parameter_group, retain_intermediates=retain_intermediates, detector_size=detector_size, 
            *args, **kwargs
        )
        self.low_mem_mode = low_memory_mode

        # This step is essential as it sets the variables to be attributes of
        # the forward modelobject. Only with this can these buffers be copied
        # to the correct devices in DataParallel.
        self.object = parameter_group.object
        self.probe = parameter_group.probe
        self.probe_positions = parameter_group.probe_positions
        self.opr_mode_weights = parameter_group.opr_mode_weights

        self.wavelength_m = wavelength_m
        self.free_space_propagation_distance_m = free_space_propagation_distance_m
        
        self.apply_subpixel_shifts_on_probe = apply_subpixel_shifts_on_probe
        self.pad_for_shift = pad_for_shift
        
        # Build free space propagator.
        self.free_space_propagator = None
        self.build_free_space_propagator()

        # Build in-object (multislice) propagator.
        self.in_object_propagator = None
        self.in_object_prop_params = None
        self.build_in_object_propagator()

        # Intermediate variables. Only used if retain_intermediate is True.
        self.intermediate_variables = self.PlanarPtychographyIntermediateVariables()

        self.check_inputs()

    def check_inputs(self):
        if self.probe.has_multiple_opr_modes:
            if self.opr_mode_weights is None:
                raise ValueError(
                    "OPRModeWeights must be given when the probe has multiple OPR modes."
                )
        if self.object.is_multislice:
            if self.wavelength_m is None:
                raise ValueError("Wavelength must be given when the object is multislice.")
            
        if self.free_space_propagation_distance_m < torch.inf and not self.pad_for_shift:
            logger.warning(
                "It seems that you are running near-field propagation "
                f"(free_space_propagation_distance_m = {self.free_space_propagation_distance_m}). "
                "If you are using a bright-field probe, it is recommended to set `pad_for_shift` "
                "to a positive integer to avoid aliasing. Currently it is 0."
            )
        
        if not self.apply_subpixel_shifts_on_probe and not self.pad_for_shift:
            ValueError(
                "Using `pad_for_shift == 0` with `apply_subpixel_shifts_on_probe == False` is "
                "prohibited because this causes serious image artifacts when applying subpixel "
                "shifts to object patches. Please set `pad_for_shift` to at least 1 and try again."
            )

    def build_in_object_propagator(self):
        if not self.object.is_multislice:
            return
        self.in_object_prop_params = WavefieldPropagatorParameters.create_simple(
            wavelength_m=self.wavelength_m,
            width_px=self.probe.shape[-1],
            height_px=self.probe.shape[-2],
            pixel_width_m=self.object.pixel_size_m,
            pixel_height_m=self.object.pixel_size_m / self.object.options.pixel_size_aspect_ratio,
            propagation_distance_m=self.object.slice_spacings.data[0],
        )
        self.in_object_propagator = AngularSpectrumPropagator(self.in_object_prop_params)
        
    def build_free_space_propagator(self):
        if self.free_space_propagation_distance_m == torch.inf:
            self.free_space_propagator = FourierPropagator()
        else:
            params = WavefieldPropagatorParameters.create_simple(
                wavelength_m=self.wavelength_m,
                width_px=self.probe.shape[-1],
                height_px=self.probe.shape[-2],
                pixel_width_m=self.object.pixel_size_m,
                pixel_height_m=self.object.pixel_size_m,
                propagation_distance_m=self.free_space_propagation_distance_m,
            )
            # TODO: AngularSpectrumPropagator uses analytical transfer function. Using the FFT
            # of real-space impulse response might offer better sampling for large propagation distances.
            if params.is_fresnel_transform_preferrable():
                self.free_space_propagator = FresnelTransformPropagator(params)
            else:
                self.free_space_propagator = AngularSpectrumPropagator(params)
                
    def extract_object_patches(self, indices: Tensor) -> Tensor:
        positions = self.probe_positions.data[indices]
        if self.apply_subpixel_shifts_on_probe:
            obj_patches = self.object.extract_patches(
                positions.round().int(), self.probe.get_spatial_shape(),
                integer_mode=True
            )
        else:
            obj_patches = self.object.extract_patches(
                positions, self.probe.get_spatial_shape(),
                pad_for_shift=self.pad_for_shift,
                integer_mode=False
            )
        return obj_patches

    def get_unique_probes(self, indices: Tensor, always_return_probe_batch: bool = True) -> Tensor:
        """Get the unique probes for all positions in the batch. 
        If OPR modes are present, this will return the unique
        probe for all positions.

        Parameters
        ----------
        indices : Tensor
            Indices of diffraction pattern in the batch.
        always_return_probe_batch : bool
            If True, the probe is always returned as a (batch_size, n_modes, h, w) tensor even if
            there is only one OPR mode. Otherwise, the probe is returned as a (n_modes, h, w) 
            tensor if there is only one OPR mode.

        Returns
        -------
        Tensor
            A (batch_size, n_modes, h, w) tensor of unique probes.
        """
        if self.probe.has_multiple_opr_modes:
            # Shape of probe:     (batch_size, n_modes, h, w)
            probe = self.probe.get_unique_probes(
                self.opr_mode_weights.get_weights(indices), mode_to_apply=0
            )
        else:
            if always_return_probe_batch:
                probe = self.probe.data.repeat(indices.shape[0], 1, 1, 1)
            else:
                probe = self.probe.get_opr_mode(0)
        return probe
    
    def shift_unique_probes(self, indices: Tensor, unique_probes: Tensor, first_mode_only: bool = False):
        """Apply subpixel shifts to the unique probes to compensate for the fractional
        probe positions not accounted for when extracting object patches.
        
        Parameters
        ----------
        indices : Tensor
            Indices of diffraction pattern in the batch.
        unique_probes : Tensor
            A (batch_size, n_modes, h, w) tensor of unique probes.
        first_mode_only : bool
            If True, only the first mode is shifted.
        """
        orig_shape = unique_probes.shape
        fractional_shifts = self.probe_positions.data[indices] - self.probe_positions.data[indices].round()
        
        if first_mode_only:
            unique_probe_to_shift = unique_probes[..., 0, :, :]
        else:
            unique_probe_to_shift = unique_probes.reshape(unique_probes.shape[0] * unique_probes.shape[1], *unique_probes.shape[2:])

        unique_probe_shifted = ip.shift_images(
            unique_probe_to_shift, 
            fractional_shifts, 
            method=self.parameter_group.object.options.patch_interpolation_method,
            adjoint=False,
            pad=self.pad_for_shift
        )
        
        if first_mode_only:
            unique_probes = torch.cat([unique_probe_shifted[..., None, :, :], unique_probes[..., 1:, :, :]], dim=-3)
        else:
            unique_probes = unique_probe_shifted.reshape(orig_shape)
        return unique_probes
        
    @timer()
    def propagate_through_object(self, probe, obj_patches):
        """
        Propagate through the planar object.

        Parameters
        ----------
        probe : Tensor
            A (batch_size, n_modes, h, w) or (n_modes, h, w) tensor
            of wavefields at the entering plane.
        obj_patches : Tensor
            A (batch_size, n_slices, h, w) tensor of object patches.

        Returns
        -------
        Tensor
            A (batch_size, n_modes, h, w) tensor of wavefields at the
            exiting plane.
        """
        slice_psi = probe
        for i_slice in range(self.parameter_group.object.n_slices):
            if i_slice > 0:
                self.record_intermediate_variable("shifted_unique_probes", slice_psi)

            # Modulate wavefield.
            # Shape of slice_psi: (batch_size, n_modes, h, w)
            slice_patches = obj_patches[:, i_slice, ...]
            slice_psi = slice_patches[:, None, :, :] * slice_psi

            # Propagate wavefield.
            if i_slice < self.parameter_group.object.n_slices - 1:
                slice_psi = self.propagate_to_next_slice(slice_psi, i_slice)
        self.record_intermediate_variable("psi", slice_psi)
        return slice_psi

    @timer()
    def forward_real_space(self, indices, obj_patches):
        """
        Propagate through the planar object.

        If `retain_intermediates` is True, the modulated and then propagated
        wavefield at each slice will also be stored in the intermediate
        variable dictionary as "slice_psis".

        Parameters
        ----------
        indices : Tensor
            Indices of diffraction pattern in the batch.
        obj_patches : Tensor
            A (batch_size, n_slices, h, w) tensor of object patches.

        Returns
        -------
        Tensor
            A (batch_size, n_modes, h, w) tensor of wavefields at the
            exiting plane.
        """
        # Shape of obj_patches:   (batch_size, n_slices, h, w)
        probe = self.get_unique_probes(indices, always_return_probe_batch=self.apply_subpixel_shifts_on_probe)
        if self.apply_subpixel_shifts_on_probe:
            probe = self.shift_unique_probes(indices, probe, first_mode_only=True)
        self.record_intermediate_variable("shifted_unique_probes", probe)
        slice_psi = self.propagate_through_object(probe, obj_patches)
        return slice_psi

    @timer()
    def forward_far_field(self, psi: Tensor) -> Tensor:
        """
        Propagate exit waves to far field.

        Parameters
        ----------
        psi: Tensor
            A (batch_size, n_probe_modes, h, w) tensor of exit waves.

        Returns
        -------
        Tensor
            A (batch_size, n_probe_modes, h, w) tensor of far field waves.
        """
        psi_far = self.free_space_propagator.propagate_forward(psi)
        self.record_intermediate_variable("psi_far", psi_far)
        return psi_far

    @timer()
    def propagate_to_next_slice(self, psi: Tensor, slice_index: int):
        """
        Propagate wavefield to the next slice by the distance given by
        `self.object.slice_spacings.data[slice_index]`.

        Parameters
        ----------
        psi : Tensor.
            A (batch_size, n_modes, h, w) complex tensor giving the wavefield at
            the current slice.
        slice_index : int
            The index of the current slice.

        Returns
        -------
        Tensor
            The wavefield propagated to the next slice.
        """
        self.in_object_prop_params = WavefieldPropagatorParameters.create_simple(
            wavelength_m=self.wavelength_m,
            width_px=self.probe.shape[-1],
            height_px=self.probe.shape[-2],
            pixel_width_m=self.object.pixel_size_m,
            pixel_height_m=self.object.pixel_size_m / self.object.options.pixel_size_aspect_ratio,
            propagation_distance_m=self.object.slice_spacings.data[slice_index],
        )
        self.in_object_propagator.update(self.in_object_prop_params)
        slice_psi_prop = self.in_object_propagator.propagate_forward(psi)
        return slice_psi_prop

    @timer()
    def propagate_to_previous_slice(self, psi: Tensor, slice_index: int):
        """
        Propagate wavefield to the previous slice by the distance given by
        `self.object.slice_spacings.data[slice_index - 1]`.

        Parameters
        ----------
        psi : Tensor.
            A (batch_size, n_modes, h, w) complex tensor giving the wavefield at
            the current slice.
        slice_index : int
            The index of the current slice.

        Returns
        -------
        Tensor
            The wavefield propagated to the next slice.
        """
        if slice_index == 0:
            return psi
        self.in_object_prop_params = WavefieldPropagatorParameters.create_simple(
            wavelength_m=self.wavelength_m,
            width_px=self.probe.shape[-1],
            height_px=self.probe.shape[-2],
            pixel_width_m=self.object.pixel_size_m,
            pixel_height_m=self.object.pixel_size_m / self.object.options.pixel_size_aspect_ratio,
            propagation_distance_m=self.object.slice_spacings.data[slice_index - 1],
        )
        self.in_object_propagator.update(self.in_object_prop_params)
        slice_psi_prop = self.in_object_propagator.propagate_backward(psi)
        return slice_psi_prop
    
    def dip_generate(self):
        """Run embedded NN models in deep image prior parameters to generate data.
        """
        if isinstance(self.object, ptychi.data_structures.object.DIPObject):
            self.object.generate()
        if isinstance(self.probe, ptychi.data_structures.probe.DIPProbe):
            self.probe.generate()
        elif isinstance(self.probe, ptychi.data_structures.probe.SynthesisDictLearnProbe):
            self.probe.generate()

    @timer()
    def forward(self, indices: Tensor, return_object_patches: bool = False) -> Tensor:
        """
        Run ptychographic forward simulation and calculate the measured intensities.

        Parameters
        ----------
        indices : Tensor
            A (N,) tensor of diffraction pattern indices in the batch.
        return_object_patches : bool
            If True, return the object patches along with the intensities.

        Returns
        -------
        Tensor
            Predicted intensities (squared magnitudes).
        """
        indices = indices.to(self.object.tensor.data.device)
        self.intermediate_variables = self.PlanarPtychographyIntermediateVariables()
        
        self.dip_generate()
        
        positions = self.probe_positions.tensor[indices]
        obj_patches = self.extract_object_patches(indices)

        self.record_intermediate_variable("positions", positions)
        self.record_intermediate_variable("obj_patches", obj_patches)

        psi = self.forward_real_space(indices, obj_patches)
        psi_far = self.forward_far_field(psi)

        y = torch.abs(psi_far) ** 2
        # Sum along probe modes
        y = y.sum(1)
        
        y = self.conform_to_detector_size(y)

        returns = [y]
        if return_object_patches:
            returns.append(obj_patches)
        if len(returns) == 1:
            return returns[0]
        else:
            return returns

    @timer()        
    def conform_to_detector_size(self, y: Tensor) -> Tensor:
        if (
            self.detector_size is None
        ) or (
            self.detector_size[0] == y.shape[-2] and self.detector_size[1] == y.shape[-1]
        ):
            return y
        if self.detector_size[0] > y.shape[-2] or self.detector_size[1] > y.shape[-1]:
            raise ValueError(
                f"Detector size ({self.detector_size}) should not be larger than the probe size ({y.shape[-2:]})."
            )
        y = torch.fft.fftshift(y, dim=(-2, -1))
        y = ip.central_crop(y, self.detector_size)
        y = torch.fft.ifftshift(y, dim=(-2, -1))
        return y

    @timer()
    def forward_low_memory(self, indices: Tensor, return_object_patches: bool = False) -> Tensor:
        """
        The forward model that should give the same result as `forward`, but
        uses less memory.

        Parameters
        ----------
        indices : Tensor
            A (N,) tensor of diffraction pattern indices in the batch.
        return_object_patches : bool
            If True, return the object patches along with the intensities.

        Returns
        -------
        Tensor
            Predicted intensities (squared magnitudes).
        """
        self.dip_generate()

        positions = self.probe_positions.tensor[indices]
        obj_patches = self.extract_object_patches(indices)

        self.record_intermediate_variable("positions", positions)
        self.record_intermediate_variable("obj_patches", obj_patches)

        probe = self.get_unique_probes(indices, always_return_probe_batch=self.apply_subpixel_shifts_on_probe)
        if self.apply_subpixel_shifts_on_probe:
            probe = self.shift_unique_probes(indices, probe, first_mode_only=True)
            self.record_intermediate_variable("shifted_unique_probes", probe)
        y = 0.0
        for i_mode in range(self.probe.n_modes):
            exit_psi = self.propagate_through_object(
                probe[..., i_mode : i_mode + 1, :, :], obj_patches
            )
            
            psi_far = self.free_space_propagator.propagate_forward(exit_psi)
            self.record_intermediate_variable("psi_far", psi_far)
            
            y = y + psi_far[..., 0, :, :].abs() ** 2

        y = self.conform_to_detector_size(y)

        # `self.intermediate_variables.shifted_unique_probes` is now a list of `n_modes * n_slices`
        # tensors. We concatenate the modes for each slice so that it becomes a list of `n_slices`
        # tensors.
        self.intermediate_variables.shifted_unique_probes = [
            torch.concat(self.intermediate_variables.shifted_unique_probes[i_mode::self.probe.n_modes], dim=-3)
            for i_mode in range(self.probe.n_modes)
        ]
        
        returns = [y]
        if return_object_patches:
            returns.append(obj_patches)
        if len(returns) == 1:
            return returns[0]
        else:
            return returns

    @timer()
    def post_differentiation_hook(self, indices, y_true, **kwargs):
        self.compensate_for_fft_scaler()
        self.scale_gradients(y_true)

    @timer()        
    def compensate_for_fft_scaler(self):
        """
        Compensate for the scaling of the FFT. If the normalization of FFT is "ortho",
        nothing is done. If the normalization is "backward" (default in `torch.fft`),
        all gradients are scaled by 1 / sqrt(N).
        """
        if (
            isinstance(self.object, ptychi.data_structures.object.DIPObject) or 
            isinstance(self.probe, ptychi.data_structures.probe.DIPProbe)
        ):
            return
        if not isinstance(self.free_space_propagator, FourierPropagator):
            return
        if self.free_space_propagator.norm == "backward" or self.free_space_propagator.norm is None:
            factor = self.probe.shape[-2] * self.probe.shape[-1]
            for var in self.optimizable_parameters:
                if var.get_grad() is not None:
                    var.set_grad(var.get_grad() / factor)

    @timer()
    def scale_gradients(self, patterns):
        """
        Scale the gradients of object and probe so that they are identical to the
        update functions of ePIE.

        For object, the ePIE update function is

            o = o + alpha * p.conj() / (abs(p) ** 2).max() * (psi_prime - psi)

        while the gradient given by AD when using MSELoss(reduction="mean") is

            -(1 / (batch_size * h * w)) * alpha * p.conj() * (psi_prime - psi)

        To scale the AD gradient to match ePIE, we should
        (1) multiply it by batch_size * h * w;
        (2) divide it by (abs(p) ** 2).max() to make up the ePIE scaling factor.

        For probe, the ePIE update function is

            p = p + alpha * mean(o.conj() / (abs(o) ** 2).max() * (psi_prime - psi), axis=0)

        while the gradient given by AD when using MSELoss(reduction="mean") is

            -(1 / (batch_size * h * w)) * alpha * sum(o.conj() * (psi_prime - psi), axis=0)

        To scale the AD gradient to match ePIE, we should
        (1) multiply it by batch_size * h * w;
        (2) divide it by (abs(o) ** 2).max() to make up the ePIE scaling factor
            (but we can assume this is 1.0);
        (3) divide it by batch_size to make up the mean over the batch dimension.
        
        For probe positions and OPR weights, we just compensate for the mean reduction:
        (1) multiply it by h * w.
        """
        if (
            isinstance(self.object, ptychi.data_structures.object.DIPObject) or 
            isinstance(self.probe, ptychi.data_structures.probe.DIPProbe)
        ):
            return
        # Directly modify the gradients here. Tensor.register_hook has memory leak issue.
        if self.object.optimizable:
            self.object.set_grad(
                self.object.get_grad()
                / self.probe.get_all_mode_intensity().max()
                * patterns.numel()
            )
        # Assuming (obj_patches.abs() ** 2).max() == 1.0
        if self.probe.optimizable:
            self.probe.set_grad(
                self.probe.get_grad() * (
                    patterns.numel() / len(patterns)
                )
            )
            
        if self.probe_positions.optimizable:
            self.probe_positions.set_grad(
                self.probe_positions.get_grad() * (
                    patterns.numel() / len(patterns)
                )
            )
            
        if self.opr_mode_weights.optimizable:
            self.opr_mode_weights.set_grad(
                self.opr_mode_weights.get_grad() * (
                    patterns.numel() / len(patterns)
                )
            )


class NoiseModel(torch.nn.Module):
    def __init__(self, eps=1e-6, valid_pixel_mask: Optional[Tensor] = None) -> None:
        super().__init__()
        self.eps = eps
        self.noise_statistics = None
        self.valid_pixel_mask = valid_pixel_mask

    def nll(self, y_pred: Tensor, y_true: Tensor) -> Tensor:
        """
        Calculate the negative log-likelihood.
        """
        raise NotImplementedError

    def backward(self, *args, **kwargs):
        raise NotImplementedError

    @timer()
    def conform_to_exit_wave_size(
        self, 
        y_pred: Tensor, 
        y_true: Tensor, 
        valid_pixel_mask: Tensor,
        psi_shape: tuple[int, int], 
    ) -> tuple[Tensor, Tensor, Tensor]:
        if psi_shape[-2] != y_pred.shape[-2] or psi_shape[-1] != y_pred.shape[-1]:
            y_pred = torch.fft.fftshift(y_pred, dim=(-2, -1))
            y_true = torch.fft.fftshift(y_true, dim=(-2, -1))
            y_pred = ip.central_crop_or_pad(y_pred, psi_shape[-2:])
            y_true = ip.central_crop_or_pad(y_true, psi_shape[-2:])
            y_pred = torch.fft.ifftshift(y_pred, dim=(-2, -1))
            y_true = torch.fft.ifftshift(y_true, dim=(-2, -1))
            if valid_pixel_mask is not None:
                valid_pixel_mask = torch.fft.fftshift(valid_pixel_mask, dim=(-2, -1))
                valid_pixel_mask = ip.central_crop_or_pad(valid_pixel_mask, psi_shape[-2:])
                valid_pixel_mask = torch.fft.ifftshift(valid_pixel_mask, dim=(-2, -1))
        return y_pred, y_true, valid_pixel_mask


class GaussianNoiseModel(NoiseModel):
    def __init__(self, sigma: float = 0.5, eps: float = 1e-9, *args, **kwargs) -> None:
        super().__init__(eps=eps, *args, **kwargs)
        self.noise_statistics = "gaussian"
        self.sigma = sigma
        self.loss_function = MSELossOfSqrt()

    def nll(self, y_pred: Tensor, y_true: Tensor) -> Tensor:
        # This is averaged over all pixels, so it differs from Eq. 11a in Odstrcil (2018)
        # by a factor of 1 / y_pred.numel().
        loss = self.loss_function(y_pred, y_true) / self.sigma**2 * 0.5
        return loss


class PtychographyGaussianNoiseModel(GaussianNoiseModel):
    def __init__(self, sigma: float = 0.5, eps: float = 1e-9, *args, **kwargs) -> None:
        super().__init__(sigma=sigma, eps=eps, *args, **kwargs)

    @timer()
    def backward_to_psi_far(self, y_pred, y_true, psi_far):
        """
        Compute the gradient of the NLL with respect to far field wavefront.
        """
        # Shape of g:       (batch_size, h, w)
        # Shape of psi_far: (batch_size, n_probe_modes, h, w)
        y_pred, y_true, valid_pixel_mask = self.conform_to_exit_wave_size(
            y_pred, y_true, self.valid_pixel_mask, psi_far.shape[-2:]
        )
        g = 1 - torch.sqrt(y_true) / (torch.sqrt(y_pred) + self.eps)  # Eq. 12b
        if valid_pixel_mask is not None:
            g[:, torch.logical_not(valid_pixel_mask)] = 0
        w = 1 / (2 * self.sigma) ** 2
        g = 2 * w * g[:, None, :, :] * psi_far
        return g


class PoissonNoiseModel(NoiseModel):
    def __init__(self, eps: float = 1e-6, *args, **kwargs) -> None:
        super().__init__(eps=eps, *args, **kwargs)
        self.noise_statistics = "poisson"
        self.loss_function = torch.nn.PoissonNLLLoss(log_input=False)

    def nll(self, y_pred: Tensor, y_true: Tensor) -> Tensor:
        # This is averaged over all pixels, so it differs from Eq. 11a in Odstrcil (2018)
        # by a factor of 1 / y_pred.numel().
        loss = self.loss_function(y_pred, y_true)
        return loss


class PtychographyPoissonNoiseModel(PoissonNoiseModel):
    def __init__(self, eps: float = 1e-6, *args, **kwargs) -> None:
        super().__init__(eps=eps, *args, **kwargs)

    @timer()
    def backward_to_psi_far(self, y_pred: Tensor, y_true: Tensor, psi_far: Tensor):
        """
        Compute the gradient of the NLL with respect to far field wavefront.
        """
        y_pred, y_true, valid_pixel_mask = self.conform_to_exit_wave_size(
            y_pred, y_true, self.valid_pixel_mask, psi_far.shape[-2:]
        )
        g = 1 - y_true / (y_pred + self.eps)  # Eq. 12b
        if valid_pixel_mask is not None:
            g[:, torch.logical_not(valid_pixel_mask)] = 0
        g = g[:, None, :, :] * psi_far
        return g
